from __future__ import annotations

from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Literal,
    overload,
)

from arro3.core import RecordBatchReader
from arro3.core.types import ArrowArrayExportable, ArrowStreamExportable

from deltalake._internal import write_to_deltalake as write_deltalake_rust
from deltalake.writer._conversion import _convert_arro3_schema_to_delta
from deltalake.writer._utils import try_get_table_and_table_uri

if TYPE_CHECKING:
    from deltalake.table import DeltaTable, WriterProperties
    from deltalake.transaction import (
        CommitProperties,
        PostCommitHookProperties,
    )


@overload
def write_deltalake(
    table_or_uri: str | Path | DeltaTable,
    data: ArrowStreamExportable | ArrowArrayExportable | Sequence[ArrowArrayExportable],
    *,
    partition_by: list[str] | str | None = ...,
    mode: Literal["error", "append", "ignore"] = ...,
    name: str | None = ...,
    description: str | None = ...,
    configuration: Mapping[str, str | None] | None = ...,
    schema_mode: Literal["merge", "overwrite"] | None = ...,
    storage_options: dict[str, str] | None = ...,
    target_file_size: int | None = ...,
    writer_properties: WriterProperties = ...,
    post_commithook_properties: PostCommitHookProperties | None = ...,
    commit_properties: CommitProperties | None = ...,
) -> None: ...


@overload
def write_deltalake(
    table_or_uri: str | Path | DeltaTable,
    data: ArrowStreamExportable | ArrowArrayExportable | Sequence[ArrowArrayExportable],
    *,
    partition_by: list[str] | str | None = ...,
    mode: Literal["overwrite"],
    name: str | None = ...,
    description: str | None = ...,
    configuration: Mapping[str, str | None] | None = ...,
    schema_mode: Literal["merge", "overwrite"] | None = ...,
    storage_options: dict[str, str] | None = ...,
    predicate: str | None = ...,
    target_file_size: int | None = ...,
    writer_properties: WriterProperties = ...,
    post_commithook_properties: PostCommitHookProperties | None = ...,
    commit_properties: CommitProperties | None = ...,
) -> None: ...


def write_deltalake(
    table_or_uri: str | Path | DeltaTable,
    data: ArrowStreamExportable | ArrowArrayExportable | Sequence[ArrowArrayExportable],
    *,
    partition_by: list[str] | str | None = None,
    mode: Literal["error", "append", "overwrite", "ignore"] = "error",
    name: str | None = None,
    description: str | None = None,
    configuration: Mapping[str, str | None] | None = None,
    schema_mode: Literal["merge", "overwrite"] | None = None,
    storage_options: dict[str, str] | None = None,
    predicate: str | None = None,
    target_file_size: int | None = None,
    writer_properties: WriterProperties | None = None,
    post_commithook_properties: PostCommitHookProperties | None = None,
    commit_properties: CommitProperties | None = None,
) -> None:
    """Write to a Delta Lake table

    If the table does not already exist, it will be created.

    Args:
        table_or_uri: URI of a table or a DeltaTable object.
        data: Data to write. If passing iterable, the schema must also be given.
        partition_by: List of columns to partition the table by. Only required
            when creating a new table.
        mode: How to handle existing data. Default is to error if table already exists.
            If 'append', will add new data.
            If 'overwrite', will replace table with new data.
            If 'ignore', will not write anything if table already exists.
        name: User-provided identifier for this table.
        description: User-provided description for this table.
        configuration: A map containing configuration options for the metadata action.
        schema_mode: If set to "overwrite", allows replacing the schema of the table. Set to "merge" to merge with existing schema.
        storage_options: options passed to the native delta filesystem.
        predicate: When using `Overwrite` mode, replace data that matches a predicate. Only used in rust engine.'
        target_file_size: Override for target file size for data files written to the delta table. If not passed, it's taken from `delta.targetFileSize`.
        writer_properties: Pass writer properties to the Rust parquet writer.
        post_commithook_properties: properties for the post commit hook. If None, default values are used.
        commit_properties: properties of the transaction commit. If None, default values are used.
    """
    table, table_uri = try_get_table_and_table_uri(table_or_uri, storage_options)
    if table is not None:
        storage_options = table._storage_options or {}
        storage_options.update(storage_options or {})
        table.update_incremental()

    if isinstance(partition_by, str):
        partition_by = [partition_by]

    if table is not None and mode == "ignore":
        return

    if isinstance(data, Sequence):
        data = RecordBatchReader.from_batches(data[0], data)  # type: ignore
    else:
        data = RecordBatchReader.from_arrow(data)

    compatible_delta_schema = _convert_arro3_schema_to_delta(data.schema)

    if table:
        table._table.write(
            data=data,
            batch_schema=compatible_delta_schema,
            partition_by=partition_by,
            mode=mode,
            schema_mode=schema_mode,
            predicate=predicate,
            target_file_size=target_file_size,
            name=name,
            description=description,
            configuration=configuration,
            writer_properties=writer_properties,
            commit_properties=commit_properties,
            post_commithook_properties=post_commithook_properties,
        )
    else:
        write_deltalake_rust(
            table_uri=table_uri,
            data=data,
            batch_schema=compatible_delta_schema,
            partition_by=partition_by,
            mode=mode,
            schema_mode=schema_mode,
            predicate=predicate,
            target_file_size=target_file_size,
            name=name,
            description=description,
            configuration=configuration,
            storage_options=storage_options,
            writer_properties=writer_properties,
            commit_properties=commit_properties,
            post_commithook_properties=post_commithook_properties,
        )



import pytest
from arro3.core import DataType, Field, Schema

from deltalake.writer._conversion import _convert_arro3_schema_to_delta


@pytest.mark.parametrize(
    "input_schema,expected_schema",
    [
        # Basic types - identity
        (
            Schema(fields=[Field("foo", DataType.int64())]),
            Schema(fields=[Field("foo", DataType.int64())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.int32())]),
            Schema(fields=[Field("foo", DataType.int32())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.int16())]),
            Schema(fields=[Field("foo", DataType.int16())]),
        ),
        # Unsigned integers to signed
        (
            Schema(fields=[Field("foo", DataType.uint8())]),
            Schema(fields=[Field("foo", DataType.int8())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.uint16())]),
            Schema(fields=[Field("foo", DataType.int16())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.uint32())]),
            Schema(fields=[Field("foo", DataType.int32())]),
        ),
        (
            Schema(fields=[Field("foo", DataType.uint64())]),
            Schema(fields=[Field("foo", DataType.int64())]),
        ),
        # Timestamps
        (
            Schema(fields=[Field("foo", DataType.timestamp("s"))]),
            Schema(fields=[Field("foo", DataType.timestamp("us"))]),
        ),
        (
            Schema(
                fields=[Field("foo", DataType.timestamp("ns", tz="Europe/Amsterdam"))]
            ),
            Schema(fields=[Field("foo", DataType.timestamp("us", tz="UTC"))]),
        ),
        # Nullability variations
        (
            Schema(fields=[Field("foo", DataType.uint16(), nullable=False)]),
            Schema(fields=[Field("foo", DataType.int16(), nullable=False)]),
        ),
        (
            Schema(fields=[Field("foo", DataType.timestamp("ns"), nullable=True)]),
            Schema(fields=[Field("foo", DataType.timestamp("us"), nullable=True)]),
        ),
        # List of unsigned ints
        (
            Schema(fields=[Field("foo", DataType.list(DataType.uint32()))]),
            Schema(fields=[Field("foo", DataType.list(DataType.int32()))]),
        ),
        (
            Schema(fields=[Field("foo", DataType.large_list(DataType.uint8()))]),
            Schema(fields=[Field("foo", DataType.large_list(DataType.int8()))]),
        ),
        # List with nullable inner fields
        (
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.uint8(), nullable=True)),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.int8(), nullable=True)),
                    )
                ]
            ),
        ),
        # List with non-nullable inner fields
        (
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.uint8(), nullable=False)),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "my_list",
                        DataType.list(Field("foo", DataType.int8(), nullable=False)),
                    )
                ]
            ),
        ),
        # Deeply nested list
        (
            Schema(
                fields=[
                    Field(
                        "deep_list",
                        DataType.list(
                            Field(
                                "level_1",
                                DataType.list(
                                    Field(
                                        "level_2",
                                        DataType.list(
                                            Field(
                                                "value",
                                                DataType.uint16(),
                                                nullable=True,
                                            )
                                        ),
                                        nullable=True,
                                    )
                                ),
                                nullable=True,
                            )
                        ),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "deep_list",
                        DataType.list(
                            Field(
                                "level_1",
                                DataType.list(
                                    Field(
                                        "level_2",
                                        DataType.list(
                                            Field(
                                                "value", DataType.int16(), nullable=True
                                            )
                                        ),
                                        nullable=True,
                                    )
                                ),
                                nullable=True,
                            )
                        ),
                    )
                ]
            ),
        ),
        # Fixed-size list
        (
            Schema(fields=[Field("foo", DataType.list(DataType.uint16(), 5))]),
            Schema(fields=[Field("foo", DataType.list(DataType.int16(), 5))]),
        ),
        # Struct with mixed fields
        (
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.struct(
                            [
                                Field("a", DataType.uint64()),
                                Field("b", DataType.timestamp("ns")),
                                Field("c", DataType.uint32()),
                            ]
                        ),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.struct(
                            [
                                Field("a", DataType.int64()),
                                Field("b", DataType.timestamp("us")),
                                Field("c", DataType.int32()),
                            ]
                        ),
                    )
                ]
            ),
        ),
        # Nested struct in list
        (
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.list(
                            DataType.struct(
                                [
                                    Field("a", DataType.uint8(), nullable=False),
                                    Field("b", DataType.timestamp("s")),
                                ]
                            )
                        ),
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.list(
                            DataType.struct(
                                [
                                    Field("a", DataType.int8(), nullable=False),
                                    Field("b", DataType.timestamp("us")),
                                ]
                            )
                        ),
                    )
                ]
            ),
        ),
        # Mixed schema with multiple fields
        (
            Schema(
                fields=[
                    Field("a", DataType.uint16()),
                    Field("b", DataType.timestamp("ms", tz="Europe/Berlin")),
                    Field(
                        "d",
                        DataType.struct(
                            [
                                Field("x", DataType.uint32()),
                                Field("y", DataType.int64()),
                            ]
                        ),
                    ),
                ]
            ),
            Schema(
                fields=[
                    Field("a", DataType.int16()),
                    Field("b", DataType.timestamp("us", tz="UTC")),
                    Field(
                        "d",
                        DataType.struct(
                            [Field("x", DataType.int32()), Field("y", DataType.int64())]
                        ),
                    ),
                ]
            ),
        ),
        # Field metadata preservations
        (
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.uint16(),
                        metadata={"description": "an unsigned int"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "foo",
                        DataType.int16(),
                        metadata={"description": "an unsigned int"},
                    )
                ]
            ),
        ),
        (
            Schema(
                fields=[
                    Field(
                        "bar",
                        DataType.timestamp("ns"),
                        nullable=True,
                        metadata={"origin": "sensor_7"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "bar",
                        DataType.timestamp("us"),
                        nullable=True,
                        metadata={"origin": "sensor_7"},
                    )
                ]
            ),
        ),
        (
            Schema(
                fields=[
                    Field(
                        "record",
                        DataType.struct(
                            [
                                Field(
                                    "id",
                                    DataType.uint32(),
                                    metadata={"index": "primary"},
                                ),
                                Field(
                                    "value",
                                    DataType.timestamp("s"),
                                    nullable=True,
                                    metadata={"unit": "seconds"},
                                ),
                            ]
                        ),
                        metadata={"type": "event"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "record",
                        DataType.struct(
                            [
                                Field(
                                    "id",
                                    DataType.int32(),
                                    metadata={"index": "primary"},
                                ),
                                Field(
                                    "value",
                                    DataType.timestamp("us"),
                                    nullable=True,
                                    metadata={"unit": "seconds"},
                                ),
                            ]
                        ),
                        metadata={"type": "event"},
                    )
                ]
            ),
        ),
        (
            Schema(
                fields=[
                    Field(
                        "sensor_readings",
                        DataType.list(
                            Field(
                                "reading",
                                DataType.uint8(),
                                nullable=False,
                                metadata={"unit": "celsius"},
                            )
                        ),
                        metadata={"shape": "1D"},
                    )
                ]
            ),
            Schema(
                fields=[
                    Field(
                        "sensor_readings",
                        DataType.list(
                            Field(
                                "reading",
                                DataType.int8(),
                                nullable=False,
                                metadata={"unit": "celsius"},
                            )
                        ),
                        metadata={"shape": "1D"},
                    )
                ]
            ),
        ),
    ],
)
def test_schema_conversion(input_schema: Schema, expected_schema: Schema):
    assert expected_schema == _convert_arro3_schema_to_delta(input_schema)



from typing import Callable

from pyarrow.dataset import Expression

from deltalake.table import FilterType

filters_to_expression: Callable[[FilterType], Expression]
_filters_to_expression: Callable[[FilterType], Expression]



import json
import os
import pathlib
from datetime import date, datetime, timezone
from decimal import Decimal
from typing import TYPE_CHECKING

import pytest
from arro3.core import Array, ChunkedArray, DataType, RecordBatchReader, Table
from arro3.core import Field as ArrowField

from deltalake import CommitProperties, DeltaTable, Transaction, write_deltalake
from deltalake._internal import Field, PrimitiveType, Schema, StructType
from deltalake.exceptions import (
    DeltaError,
    DeltaProtocolError,
    SchemaMismatchError,
)
from deltalake.query import QueryBuilder
from deltalake.writer._utils import try_get_table_and_table_uri

if TYPE_CHECKING:
    import pyarrow as pa


@pytest.mark.skip(reason="Waiting on #570")
def test_handle_existing(
    tmp_path: pathlib.Path,
    sample_data: Table,
):
    # if uri points to a non-empty directory that isn't a delta table, error
    tmp_path
    p = tmp_path / "hello.txt"
    p.write_text("hello")

    with pytest.raises(OSError) as exception:
        write_deltalake(tmp_path, sample_data, mode="overwrite")

    assert "directory is not empty" in str(exception)


@pytest.mark.pyarrow
def test_roundtrip_basic(
    tmp_path: pathlib.Path,
    sample_data_pyarrow: "pa.Table",
):
    # Check we can create the subdirectory
    import pyarrow as pa

    tmp_path = tmp_path / "path" / "to" / "table"
    start_time = datetime.now().timestamp()
    write_deltalake(tmp_path, sample_data_pyarrow)
    end_time = datetime.now().timestamp()

    assert ("0" * 20 + ".json") in os.listdir(tmp_path / "_delta_log")

    delta_table = DeltaTable(tmp_path)
    assert pa.schema(delta_table.schema()) == sample_data_pyarrow.schema

    table = delta_table.to_pyarrow_table()
    assert table == sample_data_pyarrow

    for add_path in get_add_paths(delta_table):
        # Paths should be relative, and with no partitioning have no directories
        assert "/" not in add_path

    for action in get_add_actions(delta_table):
        path = os.path.join(tmp_path, action["path"])
        actual_size = os.path.getsize(path)
        assert actual_size == action["size"]

        modification_time = action["modificationTime"] / 1000  # convert back to seconds
        assert start_time < modification_time
        assert modification_time < end_time


@pytest.mark.parametrize("mode", ["append", "overwrite"])
def test_enforce_schema_rust_writer(existing_sample_table: DeltaTable, mode: str):
    bad_data = Table(
        {
            "x": Array(
                [1, 2, 3],
                ArrowField("x", type=DataType.int64(), nullable=False),
            ),
        }
    )

    with pytest.raises(
        SchemaMismatchError,
        match=".*Cannot cast schema, number of fields does not match.*",
    ):
        write_deltalake(
            table_or_uri=existing_sample_table,
            data=bad_data,
            mode=mode,
        )

    table_uri = existing_sample_table._table.table_uri()
    with pytest.raises(
        SchemaMismatchError,
        match=".*Cannot cast schema, number of fields does not match.*",
    ):
        write_deltalake(
            table_uri,
            bad_data,
            mode=mode,
        )


def test_update_schema(existing_sample_table: DeltaTable):
    new_data = Table(
        {
            "x": Array(
                [1, 2, 3],
                ArrowField("x", type=DataType.int64(), nullable=False),
            ),
        }
    )

    with pytest.raises(DeltaError):
        write_deltalake(
            existing_sample_table, new_data, mode="append", schema_mode="overwrite"
        )

    write_deltalake(
        existing_sample_table, new_data, mode="overwrite", schema_mode="overwrite"
    )

    read_data = (
        QueryBuilder()
        .register("tbl", existing_sample_table)
        .execute("select * from tbl")
        .read_all()
    )
    assert new_data == read_data


@pytest.mark.pyarrow
def test_merge_schema(existing_table: DeltaTable):
    import pyarrow as pa

    old_table_data = existing_table.to_pyarrow_table()
    new_data = pa.table(
        {
            "new_x": pa.array([1, 2, 3], pa.int32()),
            "new_y": pa.array([1, 2, 3], pa.int32()),
        }
    )

    write_deltalake(
        existing_table,
        new_data,
        mode="append",
        schema_mode="merge",
    )
    # adjust schema of old_table_data and new_data to match each other

    for i in range(old_table_data.num_columns):
        col = old_table_data.schema.field(i)
        new_data = new_data.add_column(i, col, pa.nulls(new_data.num_rows, col.type))

    old_table_data = old_table_data.append_column(
        pa.field("new_x", pa.int32()), pa.nulls(old_table_data.num_rows, pa.int32())
    )
    old_table_data = old_table_data.append_column(
        pa.field("new_y", pa.int32()), pa.nulls(old_table_data.num_rows, pa.int32())
    )

    # define sort order
    read_data = existing_table.to_pyarrow_table().sort_by(
        [("utf8", "ascending"), ("new_x", "ascending")]
    )
    print(repr(read_data.to_pylist()))
    concated = pa.concat_tables([old_table_data, new_data])
    print(repr(concated.to_pylist()))
    assert read_data == concated

    write_deltalake(existing_table, new_data, mode="overwrite", schema_mode="overwrite")

    assert pa.schema(existing_table.schema()) == new_data.schema


@pytest.mark.xfail(reason="update schema code to allow schemaexportable as input")
def test_overwrite_schema(existing_table: DeltaTable):
    new_data_invalid = Table(
        {
            "utf8": Array(
                [1235, 546, 5645],
                ArrowField("utf8", type=DataType.int64(), nullable=False),
            ),
            "new_x": Array(
                [1, 2, 3],
                ArrowField("new_x", type=DataType.int32(), nullable=False),
            ),
            "new_y": Array(
                [1, 2, 3],
                ArrowField("new_y", type=DataType.int32(), nullable=False),
            ),
        }
    )

    with pytest.raises(DeltaError):
        write_deltalake(
            existing_table,
            new_data_invalid,
            mode="append",
            schema_mode="overwrite",
        )

    new_data = Table(
        {
            "utf8": Array(
                ["bla", "bli", "blubb"],
                ArrowField("utf8", type=DataType.utf8(), nullable=False),
            ),
            "new_x": Array(
                [1, 2, 3],
                ArrowField("new_x", type=DataType.int32(), nullable=False),
            ),
            "new_y": Array(
                [1, 2, 3],
                ArrowField("new_y", type=DataType.int32(), nullable=False),
            ),
        }
    )

    with pytest.raises(DeltaError):
        write_deltalake(
            existing_table,
            new_data,
            mode="append",
            schema_mode="overwrite",
        )

    write_deltalake(existing_table, new_data, mode="overwrite", schema_mode="overwrite")

    assert existing_table.schema().to_arrow() == new_data.schema


def test_update_schema_rust_writer_append(existing_sample_table: DeltaTable):
    with pytest.raises(SchemaMismatchError):
        # It's illegal to do schema drift without correct schema_mode
        write_deltalake(
            existing_sample_table,
            Table(
                {
                    "x4": Array(
                        [1, 2, 3],
                        ArrowField("x4", type=DataType.int64(), nullable=False),
                    ),
                }
            ),
            mode="append",
            schema_mode=None,
        )
    with pytest.raises(DeltaError):
        write_deltalake(  # schema_mode overwrite is illegal with append
            existing_sample_table,
            Table(
                {
                    "x1": Array(
                        [1, 2, 3],
                        ArrowField("x1", type=DataType.int64(), nullable=False),
                    ),
                }
            ),
            mode="append",
            schema_mode="overwrite",
        )
    write_deltalake(
        existing_sample_table,
        Table(
            {
                "x2": Array(
                    [1, 2, 3],
                    ArrowField("x2", type=DataType.int64(), nullable=False),
                ),
            }
        ),
        mode="append",
        schema_mode="merge",
    )


@pytest.mark.pyarrow
def test_write_type_castable_types(existing_table: DeltaTable):
    write_deltalake(
        existing_table,
        Table(
            {
                "utf8": Array(
                    ["1", "2", "3"],
                    ArrowField("utf8", type=DataType.string(), nullable=False),
                ),
            }
        ),
        mode="append",
        schema_mode="merge",
    )
    with pytest.raises(
        Exception,
        match="Cast error: Cannot cast string 'hello' to value of Int8 type",
    ):
        write_deltalake(
            existing_table,
            Table(
                {
                    "int8": Array(
                        ["hello", "2", "3"],
                        ArrowField("int8", type=DataType.string(), nullable=False),
                    ),
                }
            ),
            mode="append",
            schema_mode="merge",
        )

    with pytest.raises(
        Exception,
        match="Cast error: Can't cast value 1000 to type Int8",
    ):
        write_deltalake(
            existing_table,
            Table(
                {
                    "int8": Array(
                        [1000, 100, 10],
                        ArrowField("int8", type=DataType.int32(), nullable=False),
                    ),
                }
            ),
            mode="append",
            schema_mode="merge",
        )


def test_update_schema_rust_writer_invalid(existing_sample_table: DeltaTable):
    new_data = Table(
        {
            "x5": Array(
                [1, 2, 3],
                ArrowField("x5", type=DataType.int64(), nullable=False),
            ),
        }
    )

    with pytest.raises(
        SchemaMismatchError, match="Cannot cast schema, number of fields does not match"
    ):
        write_deltalake(
            existing_sample_table,
            new_data,
            mode="overwrite",
            schema_mode=None,
        )

    write_deltalake(
        existing_sample_table,
        new_data,
        mode="overwrite",
        schema_mode="overwrite",
    )

    read_data = (
        QueryBuilder()
        .register("tbl", existing_sample_table)
        .execute("select * from tbl")
        .read_all()
    )
    assert new_data == read_data


def test_merge_schema_rust_writer_with_overwrite(tmp_path: pathlib.Path):
    data = Table(
        {
            "a": Array(
                [1, 2, 3, 4],
                ArrowField("a", type=DataType.int64(), nullable=True),
            ),
            "b": Array(
                [1, 1, 2, 2],
                ArrowField("b", type=DataType.int64(), nullable=True),
            ),
            "c": Array(
                [10, 11, 12, 13],
                ArrowField("c", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data,
    )

    new_data = Table(
        {
            "a": Array(
                [100, 200, 300],
                ArrowField("a", type=DataType.int64(), nullable=True),
            ),
            "b": Array(
                [1, 1, 1],
                ArrowField("b", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        new_data,
        mode="overwrite",
        schema_mode="merge",
    )
    dt = DeltaTable(tmp_path)
    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl")
        .read_all()
        .column_names
    )
    assert set(result) == set(["a", "b", "c"])


def test_local_path(
    tmp_path: pathlib.Path,
    sample_table: Table,
    monkeypatch,
):
    monkeypatch.chdir(tmp_path)  # Make tmp_path the working directory
    (tmp_path / "path/to/table").mkdir(parents=True)

    local_path = "./path/to/table"
    write_deltalake(local_path, sample_table)
    delta_table = DeltaTable(local_path)

    table = (
        QueryBuilder()
        .register("tbl", delta_table)
        .execute("select * from tbl")
        .read_all()
    )
    assert table == sample_table


def test_local_path_with_unsafe_rename(
    tmp_path: pathlib.Path,
    sample_table: Table,
    monkeypatch,
):
    monkeypatch.chdir(tmp_path)  # Make tmp_path the working directory
    (tmp_path / "path/to/table").mkdir(parents=True)

    local_path = "./path/to/table"
    storage_opts = {
        "allow_unsafe_rename": "true",
    }
    write_deltalake(local_path, sample_table, storage_options=storage_opts)
    delta_table = DeltaTable(local_path, storage_options=storage_opts)

    table = (
        QueryBuilder()
        .register("tbl", delta_table)
        .execute("select * from tbl")
        .read_all()
    )
    assert table == sample_table


def test_roundtrip_metadata(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(
        tmp_path,
        sample_table,
        name="test_name",
        description="test_desc",
        configuration={"delta.appendOnly": "false"},
    )

    delta_table = DeltaTable(tmp_path)

    metadata = delta_table.metadata()

    assert metadata.name == "test_name"
    assert metadata.description == "test_desc"
    assert metadata.configuration == {"delta.appendOnly": "false"}


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "column",
    [
        "utf8",
        "int64",
        "int32",
        "int16",
        "int8",
        "float32",
        "float64",
        "bool",
        # "binary",
        "date32",
        "timestamp",
    ],
)
def test_roundtrip_partitioned(
    tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table", column: str
):
    import pyarrow as pa
    import pyarrow.compute as pc

    write_deltalake(tmp_path, sample_data_pyarrow, partition_by=column)

    delta_table = DeltaTable(tmp_path)
    assert pa.schema(delta_table.schema()) == sample_data_pyarrow.schema

    table = delta_table.to_pyarrow_table()
    table = table.take(pc.sort_indices(table["int64"]))
    assert table == sample_data_pyarrow

    for add_path in get_add_paths(delta_table):
        # Paths should be relative
        assert add_path.count("/") == 1


def test_roundtrip_null_partition(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    sample_table = sample_table.add_column(
        4,
        "utf8_with_nulls",
        ChunkedArray(
            Array(
                ["a", "a", "a", "a", None],
                type=ArrowField("utf8_with_nulls", DataType.string(), nullable=True),
            )
        ),
    )

    write_deltalake(
        tmp_path,
        sample_table,
        partition_by=["utf8_with_nulls"],
    )

    delta_table = DeltaTable(tmp_path)

    table = (
        QueryBuilder()
        .register("tbl", delta_table)
        .execute("select * from tbl order by price asc")
        .read_all()
    )
    assert table == sample_table


def test_roundtrip_multi_partitioned(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    write_deltalake(tmp_path, sample_table, partition_by=["sold", "price"])

    delta_table = DeltaTable(tmp_path)

    table = (
        QueryBuilder()
        .register("tbl", delta_table)
        .execute("select id, price, sold, deleted from tbl order by id asc")
        .read_all()
    )
    assert table == sample_table

    for add_path in get_add_paths(delta_table):
        # Paths should be relative
        assert add_path.count("/") == 2


def test_write_modes(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(
        tmp_path,
        sample_table,
    )
    assert (
        QueryBuilder()
        .register("tbl", DeltaTable(tmp_path))
        .execute("select * from tbl")
        .read_all()
        == sample_table
    )

    with pytest.raises(DeltaError):
        write_deltalake(tmp_path, sample_table, mode="error")

    write_deltalake(
        tmp_path,
        sample_table,
        mode="ignore",
    )
    assert ("0" * 19 + "1.json") not in os.listdir(tmp_path / "_delta_log")

    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
    )
    expected = RecordBatchReader.from_batches(
        sample_table.schema, [*sample_table.to_batches(), *sample_table.to_batches()]
    ).read_all()
    assert (
        QueryBuilder()
        .register("tbl", DeltaTable(tmp_path))
        .execute("select * from tbl")
        .read_all()
    ) == expected

    write_deltalake(
        tmp_path,
        sample_table,
        mode="overwrite",
    )
    assert (
        QueryBuilder()
        .register("tbl", DeltaTable(tmp_path))
        .execute("select * from tbl")
        .read_all()
    ) == sample_table


def test_append_only_should_append_only_with_the_overwrite_mode(  # Create rust equivalent rust
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    config = {"delta.appendOnly": "true"}

    write_deltalake(
        tmp_path,
        sample_table,
        mode="append",
        configuration=config,
    )

    table = DeltaTable(tmp_path)
    write_deltalake(
        table,
        sample_table,
        mode="append",
    )
    from deltalake.exceptions import CommitFailedError

    with pytest.raises(
        CommitFailedError,
        match="The transaction includes Remove action with data change but Delta table is append-only",
    ):
        write_deltalake(
            table,
            sample_table,
            mode="overwrite",
        )

    expected = RecordBatchReader.from_batches(
        sample_table.schema, [*sample_table.to_batches(), *sample_table.to_batches()]
    ).read_all()

    assert (
        QueryBuilder().register("tbl", table).execute("select * from tbl").read_all()
    ) == expected
    assert table.version() == 1


def test_writer_with_table(existing_sample_table: DeltaTable, sample_table: Table):
    write_deltalake(existing_sample_table, sample_table, mode="overwrite")
    assert (
        QueryBuilder()
        .register("tbl", existing_sample_table)
        .execute("select * from tbl")
        .read_all()
    ) == sample_table


@pytest.mark.pyarrow
def test_fails_wrong_partitioning(
    existing_table: DeltaTable,
    sample_data_pyarrow: "pa.Table",
):
    with pytest.raises(
        DeltaError,
        match='Generic error: Specified table partitioning does not match table partitioning: expected: [], got: ["int32"]',
    ):
        write_deltalake(
            existing_table,
            sample_data_pyarrow,
            mode="append",
            partition_by="int32",
        )


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_write_pandas(tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"):
    from pandas.testing import assert_frame_equal

    # When timestamp is converted to Pandas, it gets casted to ns resolution,
    # but Delta Lake schemas only support us resolution.
    sample_pandas = sample_data_pyarrow.to_pandas()
    write_deltalake(tmp_path, sample_pandas)
    delta_table = DeltaTable(tmp_path)
    df = delta_table.to_pandas()
    assert_frame_equal(df, sample_pandas)


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_to_pandas_with_types_mapper(tmp_path: pathlib.Path):
    """Test that DeltaTable.to_pandas() retains PyArrow Decimal type when using types_mapper."""
    import pandas as pd
    import pyarrow as pa

    schema = pa.schema(
        [
            ("id", pa.int32()),
            ("amount", pa.decimal128(18, 0)),
        ]
    )

    decimal_values = [Decimal("100"), Decimal("200"), Decimal("300")]

    data = pa.table(
        [
            pa.array([1, 2, 3], type=pa.int32()),
            pa.array(decimal_values, type=pa.decimal128(18, 0)),
        ],
        schema=schema,
    )

    delta_path = str(tmp_path / "delta_table")
    write_deltalake(delta_path, data)

    dt = DeltaTable(delta_path)

    def types_mapper(pa_type):
        if pa.types.is_decimal(pa_type):
            return pd.ArrowDtype(pa_type)
        return None

    df = dt.to_pandas(types_mapper=types_mapper)

    assert df.dtypes["amount"].pyarrow_dtype == pa.decimal128(18, 0), (
        "amount column should be Decimal(18, 0)"
    )


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "constructor",
    [
        lambda table: table.to_pyarrow_table(),
        lambda table: table.to_pyarrow_table().to_batches()[0],
    ],
)
def test_write_dataset_table_recordbatch(
    tmp_path: pathlib.Path,
    existing_table: DeltaTable,
    sample_data_pyarrow: "pa.Table",
    constructor,
):
    dataset = constructor(existing_table)
    write_deltalake(tmp_path, dataset, mode="overwrite")
    assert DeltaTable(tmp_path).to_pyarrow_table() == sample_data_pyarrow


def test_write_recordbatchreader(
    tmp_path: pathlib.Path,
    sample_table: Table,
):
    reader = RecordBatchReader.from_arrow(sample_table)
    write_deltalake(tmp_path, reader, mode="overwrite")
    assert (
        QueryBuilder()
        .register("tbl", DeltaTable(tmp_path))
        .execute("select * from tbl")
        .read_all()
        == sample_table
    )


def test_writer_partitioning(tmp_path: pathlib.Path):
    table = Table(
        {
            "p": Array(
                ["a=b", "hello world", "hello%20world"],
                ArrowField("p", type=DataType.string(), nullable=False),
            ),
            "x": Array(
                [0, 1, 2],
                ArrowField("x", type=DataType.int64(), nullable=False),
            ),
        }
    )

    write_deltalake(tmp_path, table)
    result = (
        QueryBuilder()
        .register("tbl", DeltaTable(tmp_path))
        .execute("select * from tbl")
        .read_all()
    )
    assert result == table


def get_log_path(table: DeltaTable) -> str:
    """Returns _delta_log path for this delta table."""
    return table._table.table_uri() + "/_delta_log/" + ("0" * 20 + ".json")


def get_add_actions(table: DeltaTable) -> list[str]:
    log_path = get_log_path(table)

    actions = []

    for line in open(log_path).readlines():
        log_entry = json.loads(line)

        if "add" in log_entry:
            actions.append(log_entry["add"])

    return actions


def get_stats(table: DeltaTable):
    actions = get_add_actions(table)
    # Should only have single add entry
    if len(actions) == 1:
        return json.loads(actions[0]["stats"])
    else:
        raise AssertionError("No add action found!")


def get_add_paths(table: DeltaTable) -> list[str]:
    return [action["path"] for action in get_add_actions(table)]


@pytest.mark.pyarrow
def test_writer_stats(existing_table: DeltaTable, sample_data_pyarrow: "pa.Table"):
    stats = get_stats(existing_table)

    assert stats["numRecords"] == sample_data_pyarrow.num_rows

    null_values = []
    for null_count in stats["nullCount"].values():
        if isinstance(null_count, dict):
            null_values.extend(list(null_count.values()))
        else:
            null_values.append(null_count)
    assert all(i == 0 for i in null_values)

    expected_mins = {
        "utf8": "0",
        "int64": 0,
        "int32": 0,
        "int16": 0,
        "int8": 0,
        "float32": -0.0,
        "float64": -0.0,
        "bool": False,
        "timestamp": "2022-01-01T00:00:00Z",
        "struct": {
            "x": 0,
            "y": "0",
        },
    }
    # PyArrow added support for decimal and date32 in 8.0.0
    expected_mins["decimal"] = 10.0
    expected_mins["date32"] = "2022-01-01"

    assert stats["minValues"] == expected_mins

    expected_maxs = {
        "utf8": "4",
        "int64": 4,
        "int32": 4,
        "int16": 4,
        "int8": 4,
        "float32": 4.0,
        "float64": 4.0,
        "bool": True,
        "timestamp": "2022-01-01T04:00:00Z",
        "struct": {"x": 4, "y": "4"},
    }
    # PyArrow added support for decimal and date32 in 8.0.0
    expected_maxs["decimal"] = 14.0
    expected_maxs["date32"] = "2022-01-05"

    assert stats["maxValues"] == expected_maxs


def test_writer_null_stats(tmp_path: pathlib.Path):
    data = Table(
        {
            "int32": Array(
                [1, None, 2, None],
                ArrowField("int32", type=DataType.int32(), nullable=True),
            ),
            "float64": Array(
                [1.0, None, None, None],
                ArrowField("float64", type=DataType.float64(), nullable=True),
            ),
            "str": Array(
                [None, None, None, None],
                ArrowField("str", type=DataType.string(), nullable=True),
            ),
            "bin": Array(
                [b"bindata", b"bindata", b"bindata", b"bindata"],
                ArrowField("bin", type=DataType.binary(), nullable=False),
            ),
        }
    )

    write_deltalake(tmp_path, data)

    table = DeltaTable(tmp_path)
    stats = get_stats(table)

    expected_nulls = {"int32": 2, "float64": 3, "str": 4}
    assert stats["nullCount"] == expected_nulls


def test_try_get_table_and_table_uri(tmp_path: pathlib.Path):
    def _normalize_path(t):  # who does not love Windows? ;)
        return t[0], t[1].replace("\\", "/") if t[1] else t[1]

    data = Table(
        {
            "vals": Array(
                ["1", "2", "3"],
                ArrowField("vals", type=DataType.string(), nullable=False),
            ),
        }
    )

    table_or_uri = tmp_path / "delta_table"
    write_deltalake(table_or_uri, data)
    delta_table = DeltaTable(table_or_uri)

    # table_or_uri as DeltaTable
    assert _normalize_path(
        try_get_table_and_table_uri(delta_table, None)
    ) == _normalize_path(
        (
            delta_table,
            str(tmp_path / "delta_table") + "/",
        )
    )

    # table_or_uri as str
    assert _normalize_path(
        try_get_table_and_table_uri(str(tmp_path / "delta_table"), None)
    ) == _normalize_path(
        (
            delta_table,
            str(tmp_path / "delta_table"),
        )
    )
    assert _normalize_path(
        try_get_table_and_table_uri(str(tmp_path / "str"), None)
    ) == _normalize_path(
        (
            None,
            str(tmp_path / "str"),
        )
    )

    # table_or_uri as Path
    assert _normalize_path(
        try_get_table_and_table_uri(tmp_path / "delta_table", None)
    ) == _normalize_path(
        (
            delta_table,
            str(tmp_path / "delta_table"),
        )
    )
    assert _normalize_path(
        try_get_table_and_table_uri(tmp_path / "Path", None)
    ) == _normalize_path(
        (
            None,
            str(tmp_path / "Path"),
        )
    )

    # table_or_uri with invalid parameter type
    with pytest.raises(ValueError):
        try_get_table_and_table_uri(None, None)


def param_sequence():
    import pyarrow as pa

    return (1, 2, pa.int64(), "1")


def param_sequence2():
    import pyarrow as pa

    return (False, True, pa.bool_(), "false")


def param_sequence3():
    import pyarrow as pa

    return (date(2022, 1, 1), date(2022, 1, 2), pa.date32(), "'2022-01-01'")


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "func",
    [
        param_sequence,
        param_sequence2,
        param_sequence3,
    ],
)
def test_partition_overwrite(
    tmp_path: pathlib.Path,
    func,
):
    import pyarrow as pa

    value_1, value_2, value_type, filter_string = func()
    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([1, 1, 1, 1], pa.int64()),
        }
    )
    write_deltalake(
        tmp_path, sample_data_pyarrow, mode="overwrite", partition_by=["p1", "p2"]
    )

    delta_table = DeltaTable(tmp_path)
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == sample_data_pyarrow
    )

    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1", "1"], pa.string()),
            "p2": pa.array([value_2, value_1], value_type),
            "val": pa.array([2, 2], pa.int64()),
        }
    )
    expected_data = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([2, 2, 1, 1], pa.int64()),
        }
    )
    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="overwrite",
        predicate="p1 = 1",
    )

    delta_table.update_incremental()
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == expected_data
    )

    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1", "2"], pa.string()),
            "p2": pa.array([value_2, value_2], value_type),
            "val": pa.array([3, 3], pa.int64()),
        }
    )
    expected_data = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([2, 3, 1, 3], pa.int64()),
        }
    )

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="overwrite",
        predicate=f"p2 > {filter_string}",
    )
    delta_table.update_incremental()
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == expected_data
    )

    # Overwrite a single partition
    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1"], pa.string()),
            "p2": pa.array([value_1], value_type),
            "val": pa.array([5], pa.int64()),
        }
    )
    expected_data = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([5, 3, 1, 3], pa.int64()),
        }
    )
    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="overwrite",
        predicate=f"p1 = 1 AND p2 = {filter_string}",
    )
    delta_table.update_incremental()
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == expected_data
    )
    with pytest.raises(DeltaProtocolError, match="Invariant violations"):
        write_deltalake(
            tmp_path,
            sample_data_pyarrow,
            mode="overwrite",
            predicate=f"p2 < {filter_string}",
        )


@pytest.fixture()
def sample_data_for_partitioning() -> Table:
    return Table(
        {
            "p1": Array(
                ["1", "1", "2", "2"],
                ArrowField("p1", type=DataType.string(), nullable=False),
            ),
            "p2": Array(
                [1, 2, 1, 2],
                ArrowField("p2", type=DataType.int64(), nullable=False),
            ),
            "val": Array(
                [1, 1, 1, 1],
                ArrowField("val", type=DataType.int64(), nullable=False),
            ),
        }
    )


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "func",
    [
        param_sequence,
        param_sequence2,
        param_sequence3,
    ],
)
def test_replace_where_overwrite(
    tmp_path: pathlib.Path,
    func,
):
    value_1, value_2, value_type, filter_string = func()
    import pyarrow as pa

    table_path = tmp_path

    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([1, 1, 1, 1], pa.int64()),
        }
    )
    write_data = sample_data_pyarrow

    write_deltalake(table_path, write_data, mode="overwrite")

    delta_table = DeltaTable(table_path)
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == sample_data_pyarrow
    )

    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1", "1"], pa.string()),
            "p2": pa.array([value_2, value_1], value_type),
            "val": pa.array([2, 3], pa.int64()),
        }
    )
    expected_data = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([3, 2, 1, 1], pa.int64()),
        }
    )

    write_deltalake(
        table_path,
        sample_data_pyarrow,
        mode="overwrite",
        predicate="p1 = '1'",
    )

    delta_table.update_incremental()
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == expected_data
    )


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "func",
    [
        param_sequence,
        param_sequence2,
        param_sequence3,
    ],
)
def test_replace_where_overwrite_partitioned(
    tmp_path: pathlib.Path,
    func,
):
    import pyarrow as pa

    value_1, value_2, value_type, filter_string = func()
    table_path = tmp_path

    sample_data_pyarrow = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([1, 1, 1, 1], pa.int64()),
        }
    )
    write_deltalake(
        table_path, sample_data_pyarrow, mode="overwrite", partition_by=["p1", "p2"]
    )

    delta_table = DeltaTable(table_path)
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == sample_data_pyarrow
    )

    replace_data = pa.table(
        {
            "p1": pa.array(["1", "1"], pa.string()),
            "p2": pa.array([value_2, value_1], value_type),
            "val": pa.array([2, 3], pa.int64()),
        }
    )
    expected_data = pa.table(
        {
            "p1": pa.array(["1", "1", "2", "2"], pa.string()),
            "p2": pa.array([value_1, value_2, value_1, value_2], value_type),
            "val": pa.array([3, 2, 1, 1], pa.int64()),
        }
    )

    write_deltalake(
        table_path,
        replace_data,
        mode="overwrite",
        partition_by=["p1", "p2"],
        predicate="p1 = '1'",
    )

    delta_table.update_incremental()
    assert (
        delta_table.to_pyarrow_table().sort_by(
            [("p1", "ascending"), ("p2", "ascending")]
        )
        == expected_data
    )


def test_partition_overwrite_with_new_partition(
    tmp_path: pathlib.Path, sample_data_for_partitioning: Table
):
    write_deltalake(
        tmp_path,
        sample_data_for_partitioning,
        mode="overwrite",
        partition_by=["p1", "p2"],
    )

    new_sample_data = Table(
        {
            "p1": Array(
                ["1", "2"],
                ArrowField("p1", type=DataType.string(), nullable=False),
            ),
            "p2": Array(
                [2, 2],
                ArrowField("p2", type=DataType.int64(), nullable=False),
            ),
            "val": Array(
                [2, 3],
                ArrowField("val", type=DataType.int64(), nullable=False),
            ),
        }
    )

    expected_data = Table(
        {
            "p1": Array(
                ["1", "1", "2", "2"],
                ArrowField("p1", type=DataType.string(), nullable=False),
            ),
            "p2": Array(
                [1, 2, 1, 2],
                ArrowField("p2", type=DataType.int64(), nullable=False),
            ),
            "val": Array(
                [1, 2, 1, 3],
                ArrowField("val", type=DataType.int64(), nullable=False),
            ),
        }
    )

    write_deltalake(tmp_path, new_sample_data, mode="overwrite", predicate="p2 = 2")
    delta_table = DeltaTable(tmp_path)
    result = (
        QueryBuilder()
        .register("tbl", delta_table)
        .execute("select p1,p2,val from tbl order by p1 asc, p2 asc")
        .read_all()
    )
    assert result == expected_data


def test_partition_overwrite_with_non_partitioned_data(
    tmp_path: pathlib.Path, sample_data_for_partitioning: Table
):
    write_deltalake(tmp_path, sample_data_for_partitioning, mode="overwrite")
    write_deltalake(
        tmp_path,
        Table(
            {
                "p1": Array(
                    ["1", "1"],
                    ArrowField("p1", type=DataType.string(), nullable=False),
                ),
                "p2": Array(
                    [
                        1,
                        2,
                    ],
                    ArrowField("p2", type=DataType.int64(), nullable=False),
                ),
                "val": Array(
                    [1, 1],
                    ArrowField("val", type=DataType.int64(), nullable=False),
                ),
            }
        ),
        mode="overwrite",
        predicate="p1 = 1",
    )


def test_partition_overwrite_with_wrong_partition(
    tmp_path: pathlib.Path, sample_data_for_partitioning: Table
):
    write_deltalake(
        tmp_path,
        sample_data_for_partitioning,
        mode="overwrite",
        partition_by=["p1", "p2"],
    )
    from deltalake.exceptions import DeltaError

    with pytest.raises(DeltaError, match="No field named p999."):
        write_deltalake(
            tmp_path,
            sample_data_for_partitioning,
            mode="overwrite",
            predicate="p999 = 1",
        )

    new_data = Table(
        {
            "p1": Array(
                ["1"],
                ArrowField("p1", type=DataType.string(), nullable=False),
            ),
            "p2": Array(
                [2],
                ArrowField("p2", type=DataType.int64(), nullable=False),
            ),
            "val": Array(
                [1],
                ArrowField("val", type=DataType.int64(), nullable=False),
            ),
        }
    )

    with pytest.raises(
        DeltaProtocolError,
        match="Invariant violations",
    ):
        write_deltalake(
            tmp_path,
            new_data,
            mode="overwrite",
            predicate="p1 = 1 AND p2 = 1",
        )


def test_handles_binary_data(tmp_path: pathlib.Path):
    table = Table(
        {
            "field_one": Array(
                [b"\x00\\"],
                ArrowField("field_one", type=DataType.binary(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, table)

    dt = DeltaTable(tmp_path)
    out = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
    assert table == out


@pytest.mark.pyarrow
def test_large_arrow_types(tmp_path: pathlib.Path):
    import pyarrow as pa

    pylist = [
        {"name": "Joey", "gender": b"M", "arr_type": ["x", "y"], "dict": {"a": b"M"}},
        {"name": "Ivan", "gender": b"F", "arr_type": ["x", "z"]},
    ]
    schema = pa.schema(
        [
            pa.field("name", pa.large_string()),
            pa.field("gender", pa.large_binary()),
            pa.field(
                "arr_type",
                pa.large_list(pa.field("element", pa.large_string(), nullable=False)),
            ),
            pa.field("map_type", pa.map_(pa.large_string(), pa.large_binary())),
            pa.field("struct", pa.struct([pa.field("sub", pa.large_string())])),
        ]
    )
    table = pa.Table.from_pylist(pylist, schema=schema)

    write_deltalake(tmp_path, table, mode="append")
    write_deltalake(tmp_path, table, mode="append")
    write_deltalake(tmp_path, table, mode="append")

    dt = DeltaTable(tmp_path)
    assert table.schema == pa.schema(dt.schema().to_arrow(as_large_types=True))


@pytest.mark.pyarrow
def test_large_arrow_types_dataset_as_large_types(tmp_path: pathlib.Path):
    import pyarrow as pa
    from pyarrow.dataset import dataset

    pylist = [
        {"name": "Joey", "gender": b"M", "arr_type": ["x", "y"], "dict": {"a": b"M"}},
        {"name": "Ivan", "gender": b"F", "arr_type": ["x", "z"]},
    ]
    schema = pa.schema(
        [
            pa.field("name", pa.large_string()),
            pa.field("gender", pa.large_binary()),
            pa.field(
                "arr_type",
                pa.large_list(pa.field("element", pa.large_string(), nullable=False)),
            ),
            pa.field("map_type", pa.map_(pa.large_string(), pa.large_binary())),
            pa.field("struct", pa.struct([pa.field("sub", pa.large_string())])),
        ]
    )
    table = pa.Table.from_pylist(pylist, schema=schema)

    write_deltalake(tmp_path, table)

    dt = DeltaTable(tmp_path)

    ds = dt.to_pyarrow_dataset(as_large_types=True)
    union_ds = dataset([ds, dataset(table)])
    assert union_ds.to_table().shape[0] == 4


@pytest.mark.pyarrow
def test_large_arrow_types_explicit_scan_schema(tmp_path: pathlib.Path):
    import pyarrow as pa
    from pyarrow.dataset import dataset

    pylist = [
        {"name": "Joey", "gender": b"M", "arr_type": ["x", "y"], "dict": {"a": b"M"}},
        {"name": "Ivan", "gender": b"F", "arr_type": ["x", "z"]},
    ]
    schema = pa.schema(
        [
            pa.field("name", pa.large_string()),
            pa.field("gender", pa.large_binary()),
            pa.field("arr_type", pa.large_list(pa.large_string())),
            pa.field("map_type", pa.map_(pa.large_string(), pa.large_binary())),
            pa.field("struct", pa.struct([pa.field("sub", pa.large_string())])),
        ]
    )
    table = pa.Table.from_pylist(pylist, schema=schema)

    write_deltalake(tmp_path, table)

    dt = DeltaTable(tmp_path)

    ds = dt.to_pyarrow_dataset(schema=schema)
    union_ds = dataset([ds, dataset(table)])
    assert union_ds.to_table().shape[0] == 4


def test_partition_large_arrow_types(tmp_path: pathlib.Path):
    table = Table(
        {
            "foo": Array(
                ["1", "1", "2", "2"],
                ArrowField("foo", type=DataType.large_utf8(), nullable=True),
            ),
            "bar": Array(
                [1, 2, 1, 2],
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [1, 1, 1, 1],
                ArrowField("baz", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, table, partition_by=["foo"])

    dt = DeltaTable(tmp_path)
    files = dt.files()
    expected = ["foo=1", "foo=2"]

    result = sorted([file.split("/")[0] for file in files])
    assert expected == result


def test_uint_arrow_types(tmp_path: pathlib.Path):
    table = Table(
        {
            "num1": Array(
                [3, 1],
                ArrowField("num1", type=DataType.uint8(), nullable=True),
            ),
            "num2": Array(
                [3, 13],
                ArrowField("num2", type=DataType.uint16(), nullable=True),
            ),
            "num3": Array(
                [3, 35],
                ArrowField("num3", type=DataType.uint32(), nullable=True),
            ),
            "num4": Array(
                [5, 13],
                ArrowField("num4", type=DataType.uint64(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, table)


@pytest.mark.pyarrow
def test_issue_1651_roundtrip_timestamp(tmp_path: pathlib.Path):
    import pyarrow as pa

    data = pa.table(
        {
            "id": pa.array([425], type=pa.int32()),
            "data": pa.array(["python-module-test-write"]),
            "t": pa.array([datetime(2023, 9, 15)]),
        }
    )

    write_deltalake(table_or_uri=tmp_path, mode="append", data=data, partition_by=["t"])
    dt = DeltaTable(table_uri=tmp_path)
    dataset = dt.to_pyarrow_dataset()

    assert dataset.count_rows() == 1


@pytest.mark.pyarrow
def test_invalid_decimals(tmp_path: pathlib.Path):
    import re

    import pyarrow as pa

    data = pa.table(
        {"x": pa.array([Decimal("10000000000000000000000000000000000000.0")])}
    )

    with pytest.raises(
        SchemaMismatchError,
        match=re.escape("Invalid data type for Delta Lake: Decimal256(39, 1)"),
    ):
        write_deltalake(table_or_uri=tmp_path, mode="append", data=data)


@pytest.mark.pyarrow
def test_write_large_decimal(tmp_path: pathlib.Path):
    import pyarrow as pa

    data = pa.table(
        {
            "decimal_column": pa.array(
                [Decimal(11111111111111111), Decimal(22222), Decimal("333333333333.33")]
            )
        }
    )

    write_deltalake(tmp_path, data)


def test_float_values(tmp_path: pathlib.Path):
    data = Table(
        {
            "id": Array(
                [0, 1, 2, 3],
                ArrowField("id", type=DataType.int64(), nullable=True),
            ),
            "x1": Array(
                [0.0, float("inf"), None, 1.0],
                ArrowField("x1", type=DataType.float64(), nullable=True),
            ),
            "x2": Array(
                [0.0, float("-inf"), None, 1.0],
                ArrowField("x2", type=DataType.float64(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data)
    dt = DeltaTable(tmp_path)
    assert (
        QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
        == data
    )

    actions = dt.get_add_actions()

    def get_value(name: str):
        return actions.column(name)[0].as_py()

    # x1 has no max, since inf was the highest value
    assert get_value("min")["x1"] == -0.0
    assert get_value("max")["x1"] is None
    assert get_value("null_count")["x1"] == 1
    # x2 has no min, since -inf was the lowest value
    assert get_value("min")["x2"] is None
    assert get_value("max")["x2"] == 1.0
    assert get_value("null_count")["x2"] == 1


def test_schema_cols_diff_order(tmp_path: pathlib.Path):
    data = Table(
        {
            "foo": Array(
                ["B"] * 10,
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
            "bar": Array(
                [1] * 10,
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [2.0] * 10,
                ArrowField("baz", type=DataType.float64(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="append")

    data = Table(
        {
            "baz": Array(
                [2.0] * 10,
                ArrowField("baz", type=DataType.float64(), nullable=True),
            ),
            "bar": Array(
                [1] * 10,
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "foo": Array(
                ["B"] * 10,
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="append")
    dt = DeltaTable(tmp_path)
    assert dt.version() == 1

    expected = Table(
        {
            "baz": Array(
                [2.0] * 20,
                ArrowField("baz", type=DataType.float64(), nullable=True),
            ),
            "bar": Array(
                [1] * 20,
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "foo": Array(
                ["B"] * 20,
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
        }
    )

    assert (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select baz, bar, foo from tbl")
        .read_all()
        .rechunk()
    ) == expected


@pytest.mark.pyarrow
def test_empty(existing_table: DeltaTable):
    import pyarrow as pa

    schema = existing_table.schema()
    expected = existing_table.to_pyarrow_table()
    empty_table = pa.Table.from_pylist([], schema=pa.schema(schema))
    write_deltalake(
        existing_table,
        empty_table,
        mode="append",
    )

    existing_table.update_incremental()
    assert existing_table.version() == 1
    assert expected == existing_table.to_pyarrow_table()


@pytest.mark.pyarrow
def test_rust_decimal_cast(tmp_path: pathlib.Path):
    import re

    import pyarrow as pa

    data = pa.table({"x": pa.array([Decimal("100.1")])})

    write_deltalake(
        tmp_path,
        data,
        mode="append",
    )

    assert DeltaTable(tmp_path).to_pyarrow_table()["x"][0].as_py() == Decimal("100.1")

    # Write smaller decimal,  works since it's fits in the previous decimal precision, scale
    data = pa.table({"x": pa.array([Decimal("10.1")])})
    write_deltalake(
        tmp_path,
        data,
        mode="append",
    )

    data = pa.table({"x": pa.array([Decimal("1000.1")])})
    # write decimal that is larger than target type in table
    with pytest.raises(
        SchemaMismatchError,
        match=re.escape(
            "Cannot cast field x from Decimal128(5, 1) to Decimal128(4, 1)"
        ),
    ):
        write_deltalake(
            tmp_path,
            data,
            mode="append",
        )

    with pytest.raises(SchemaMismatchError):
        write_deltalake(
            tmp_path,
            data,
            mode="append",
            schema_mode="merge",
        )


def test_write_stats_column_idx(tmp_path: pathlib.Path):
    def _check_stats(dt: DeltaTable):
        add_actions_table = dt.get_add_actions(flatten=True)

        def get_value(name: str):
            return add_actions_table.column(name)[0].as_py()

        assert get_value("null_count.foo") == 2
        assert get_value("min.foo") == "a"
        assert get_value("max.foo") == "b"
        assert get_value("null_count.bar") == 1
        assert get_value("min.bar") == 1
        assert get_value("max.bar") == 3
        assert get_value("null_count.baz") is None
        assert get_value("min.baz") is None
        assert get_value("max.baz") is None

    data = Table(
        {
            "foo": Array(
                ["a", "b", None, None],
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
            "bar": Array(
                [1, 2, 3, None],
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [1, 1, None, None],
                ArrowField("baz", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data,
        mode="append",
        configuration={"delta.dataSkippingNumIndexedCols": "2"},
    )

    dt = DeltaTable(tmp_path)
    _check_stats(dt)

    # Check if it properly takes skippingNumIndexCols from the config in the table
    write_deltalake(tmp_path, data, mode="overwrite")

    dt = DeltaTable(tmp_path)
    assert dt.version() == 1
    _check_stats(dt)


def test_write_stats_columns_stats_provided(tmp_path: pathlib.Path):
    def _check_stats(dt: DeltaTable):
        add_actions_table = dt.get_add_actions(flatten=True)

        def get_value(name: str):
            return add_actions_table.column(name)[0].as_py()

        assert get_value("null_count.foo") == 2
        assert get_value("min.foo") == "a"
        assert get_value("max.foo") == "b"
        assert get_value("null_count.bar") is None
        assert get_value("min.bar") is None
        assert get_value("max.bar") is None
        assert get_value("null_count.baz") == 2
        assert get_value("min.baz") == 1
        assert get_value("max.baz") == 1

    data = Table(
        {
            "foo": Array(
                ["a", "b", None, None],
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
            "bar": Array(
                [1, 2, 3, None],
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [1, 1, None, None],
                ArrowField("baz", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data,
        mode="append",
        configuration={"delta.dataSkippingStatsColumns": "foo,`baz`"},
    )

    dt = DeltaTable(tmp_path)
    _check_stats(dt)

    # Check if it properly takes skippingNumIndexCols from the config in the table
    write_deltalake(tmp_path, data, mode="overwrite")

    dt = DeltaTable(tmp_path)
    assert dt.version() == 1
    _check_stats(dt)


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "array",
    [
        lambda: __import__("pyarrow").array([[datetime(2010, 1, 1)]]),
        lambda: __import__("pyarrow").array([{"foo": datetime(2010, 1, 1)}]),
        lambda: __import__("pyarrow").array([{"foo": [[datetime(2010, 1, 1)]]}]),
        lambda: __import__("pyarrow").array(
            [{"foo": [[{"foo": datetime(2010, 1, 1)}]]}]
        ),
    ],
)
def test_write_timestamp_ntz_nested(tmp_path: pathlib.Path, array):
    import pyarrow as pa

    data = pa.table({"x": array()})
    write_deltalake(
        tmp_path,
        data,
        mode="append",
    )

    dt = DeltaTable(tmp_path)

    protocol = dt.protocol()
    assert protocol.min_reader_version == 3
    assert protocol.min_writer_version == 7
    assert protocol.reader_features == ["timestampNtz"]
    assert protocol.writer_features == ["timestampNtz"]


def test_parse_stats_with_new_schema(tmp_path):
    data = Table(
        {
            "val": Array(
                [1, 1],
                ArrowField("val", type=DataType.int8(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data)

    data = Table(
        {
            "val": Array(
                [1000000000000, 1000000000000],
                ArrowField("val", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="overwrite", schema_mode="overwrite")


def test_roundtrip_cdc_evolution(tmp_path: pathlib.Path):
    """
    This test is used as a CDC integration test from Python to ensure,
    approximately, that CDC files are being written
    """
    raw_commit = r"""{"metaData":{"id":"bb0fdeb2-76dd-4f5e-b1ea-845ecec8fa7e","format":{"provider":"parquet","options":{}},"schemaString":"{\"type\":\"struct\",\"fields\":[{\"name\":\"id\",\"type\":\"integer\",\"nullable\":true,\"metadata\":{}}]}","partitionColumns":[],"configuration":{"delta.enableChangeDataFeed":"true"},"createdTime":1713110303902}}
{"protocol":{"minReaderVersion":1,"minWriterVersion":4,"writerFeatures":["changeDataFeed"]}}
"""
    # timestampNtz looks like it might be an unnecessary requirement to write from Python
    os.mkdir(os.path.join(tmp_path, "_delta_log"))
    # This is a stupid hack to make sure we have a CDC capable table from the jump
    with open(
        os.path.join(tmp_path, "_delta_log", "00000000000000000000.json"), "w+"
    ) as fd:
        fd.write(raw_commit)
    assert ("0" * 20 + ".json") in os.listdir(tmp_path / "_delta_log")

    # Make sure the _change_data doesn't exist
    assert not os.path.isdir(os.path.join(tmp_path, "_change_data"))

    nrows = 5
    data = Table(
        {
            "utf8": Array(
                [str(x) for x in range(nrows)],
                ArrowField("utf8", type=DataType.utf8(), nullable=True),
            ),
            "int64": Array(
                list(range(nrows)),
                ArrowField("int64", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data,
        mode="append",
        schema_mode="merge",
    )
    assert ("0" * 19 + "1.json") in os.listdir(tmp_path / "_delta_log")

    delta_table = DeltaTable(tmp_path)
    delta_table.update(predicate="utf8 = '1'", updates={"utf8": "'hello world'"})

    delta_table = DeltaTable(tmp_path)
    print(os.listdir(tmp_path))
    # This is kind of a weak test to verify that CDFs were written
    assert os.path.isdir(os.path.join(tmp_path, "_change_data"))


@pytest.mark.pyarrow
def test_empty_data_write(tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"):
    empty_arrow_table = sample_data_pyarrow.schema.empty_table()
    write_deltalake(tmp_path, empty_arrow_table, mode="append")
    dt = DeltaTable(tmp_path)

    new_dataset = dt.to_pyarrow_dataset()
    assert new_dataset.count_rows() == 0


@pytest.mark.pandas
def test_predicate_out_of_bounds(tmp_path: pathlib.Path):
    """See <https://github.com/delta-io/delta-rs/issues/2867>"""
    import pandas as pd

    data = [
        (datetime(2024, 7, 31, 9, 30, 0), "AAPL", "20240731", 100, 11.1),
        (datetime(2024, 7, 31, 9, 30, 0), "GOOG", "20240731", 200, 11.1),
    ]
    columns = ["ts", "ins", "date", "f1", "f2"]
    df = pd.DataFrame(data, columns=columns)

    predicate = "date == 20240731"
    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        partition_by="date",
        mode="overwrite",
        schema_mode="merge",
        predicate=predicate,
    )

    data = [
        (datetime(2024, 7, 31, 9, 30, 0), "AAPL", "20240731", 666, 666),
        (datetime(2024, 7, 31, 9, 30, 0), "GOOG", "20240731", 777, 777),
    ]
    columns = ["ts", "ins", "date", "fb", "fc"]
    df = pd.DataFrame(data, columns=columns)
    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        partition_by="date",
        mode="overwrite",
        schema_mode="merge",
        predicate=predicate,
    )


@pytest.mark.pandas
def test_write_timestampntz(tmp_path: pathlib.Path):
    import pandas as pd

    data = [
        ("AAPL", "20240731", 100, 11.1),
        ("GOOG", "20240731", 200, 11.1),
    ]
    columns = ["ins", "date", "f1", "f2"]
    df = pd.DataFrame(data, columns=columns)

    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        partition_by="date",
        mode="overwrite",
    )

    dt = DeltaTable(tmp_path)
    protocol = dt.protocol()
    # A fresh table with no special features should have the lowest possible
    # minwriter feature
    assert protocol.min_writer_version == 2

    data = [
        (datetime(2024, 7, 31, 9, 30, 0), "AAPL", "20240731", 666, 666),
        (datetime(2024, 7, 31, 9, 30, 0), "GOOG", "20240731", 777, 777),
    ]
    columns = ["ts", "ins", "date", "fb", "fc"]
    df = pd.DataFrame(data, columns=columns)
    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        partition_by="date",
        mode="append",
        schema_mode="merge",
    )

    dt = DeltaTable(tmp_path)
    protocol = dt.protocol()
    # Now that a datetime has been passed through the writer version needs to
    # be upgraded to 7 to support timestampNtz
    assert protocol.min_writer_version == 7


@pytest.mark.pandas
def test_write_timestamp(tmp_path: pathlib.Path):
    import pandas as pd

    data = [
        ("AAPL", "20240731", 100, 11.1),
        ("GOOG", "20240731", 200, 11.1),
    ]
    columns = ["ins", "date", "f1", "f2"]
    df = pd.DataFrame(data, columns=columns)

    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        partition_by="date",
        mode="overwrite",
    )

    dt = DeltaTable(tmp_path)
    protocol = dt.protocol()
    # A fresh table with no special features should have the lowest possible
    # minwriter feature
    assert protocol.min_writer_version == 2

    # Performing schema evolution with a timestamp that *has* a timezone should
    # not result in a writer version upgrade!
    data = [
        (
            datetime(2024, 7, 31, 9, 30, 0, tzinfo=timezone.utc),
            "AAPL",
            "20240731",
            666,
            666,
        ),
        (
            datetime(2024, 7, 31, 9, 30, 0, tzinfo=timezone.utc),
            "GOOG",
            "20240731",
            777,
            777,
        ),
    ]
    columns = ["ts", "ins", "date", "fb", "fc"]
    df = pd.DataFrame(data, columns=columns)
    write_deltalake(
        table_or_uri=tmp_path,
        data=df,
        partition_by="date",
        mode="append",
        schema_mode="merge",
    )

    # Reload the table to make sure we have the latest protocol
    dt = DeltaTable(tmp_path)
    protocol = dt.protocol()
    # Now that a datetime has been passed through the writer version needs to
    # be upgraded to 7 to support timestampNtz
    assert protocol.min_writer_version == 2


def test_write_transactions(tmp_path: pathlib.Path, sample_table: Table):
    expected_transactions = [
        Transaction(app_id="app_1", version=1),
        Transaction(app_id="app_2", version=2, last_updated=123456),
    ]
    commit_properties = CommitProperties(app_transactions=expected_transactions)
    write_deltalake(
        table_or_uri=tmp_path,
        data=sample_table,
        mode="overwrite",
        schema_mode="overwrite",
        commit_properties=commit_properties,
    )

    delta_table = DeltaTable(tmp_path)
    transaction_1 = delta_table.transaction_version("app_1")
    assert transaction_1 == 1

    transaction_2 = delta_table.transaction_version("app_2")
    assert transaction_2 == 2


@pytest.mark.polars
def test_write_structs(tmp_path: pathlib.Path):
    import polars as pl

    dt = DeltaTable.create(
        tmp_path,
        schema=Schema(
            fields=[
                Field(name="a", type=PrimitiveType("integer")),
                Field(name="b", type=PrimitiveType("string")),
                Field(
                    name="c",
                    type=StructType(
                        [
                            Field(name="d", type=PrimitiveType("short")),
                            Field(name="e", type=PrimitiveType("short")),
                        ]
                    ),
                ),
            ]
        ),
    )

    df = pl.DataFrame(
        {
            "a": [0, 1],
            "b": ["x", "y"],
            "c": [
                {"d": -55, "e": -32},
                {"d": 0, "e": 0},
            ],
        }
    )

    dt.merge(
        source=df,
        predicate=" AND ".join([f"target.{x} = source.{x}" for x in ["a"]]),
        source_alias="source",
        target_alias="target",
    ).when_not_matched_insert_all().execute()


@pytest.mark.polars
@pytest.mark.xfail(reason="polars needs update")
def test_write_type_coercion_predicate(tmp_path: pathlib.Path):
    import polars as pl

    df = pl.DataFrame({"A": [1, 2], "B": ["hi", "hello"], "C": ["a", "b"]})
    df.write_delta(tmp_path)

    df = pl.DataFrame({"A": [10], "B": ["yeah"], "C": ["a"]})
    df.write_delta(
        tmp_path,
        mode="overwrite",
        delta_write_options=dict(predicate="C = 'a'"),
    )


@pytest.mark.pyarrow
def test_write_schema_evolved_same_metadata_id(tmp_path):
    """https://github.com/delta-io/delta-rs/issues/3274"""
    import pyarrow as pa

    data_first_write = pa.array(
        [
            {"name": "Alice", "age": 30, "details": {"email": "alice@example.com"}},
            {"name": "Bob", "age": 25, "details": {"email": "bob@example.com"}},
        ]
    )

    data_second_write = pa.array(
        [
            {
                "name": "Charlie",
                "age": 35,
                "details": {"address": "123 Main St", "email": "charlie@example.com"},
            },
            {
                "name": "Diana",
                "age": 28,
                "details": {"address": "456 Elm St", "email": "diana@example.com"},
            },
        ]
    )

    schema_first_write = pa.schema(
        [
            ("name", pa.string()),
            ("age", pa.int64()),
            ("details", pa.struct([("email", pa.string())])),
        ]
    )

    schema_second_write = pa.schema(
        [
            ("name", pa.string()),
            ("age", pa.int64()),
            (
                "details",
                pa.struct(
                    [
                        ("address", pa.string()),
                        ("email", pa.string()),
                    ]
                ),
            ),
        ]
    )
    table_first_write = pa.Table.from_pylist(
        data_first_write, schema=schema_first_write
    )
    table_second_write = pa.Table.from_pylist(
        data_second_write, schema=schema_second_write
    )

    write_deltalake(
        tmp_path,
        table_first_write,
        mode="append",
    )

    first_metadata_id = DeltaTable(tmp_path).metadata().id

    write_deltalake(tmp_path, table_second_write, mode="append", schema_mode="merge")

    second_metadata_id = DeltaTable(tmp_path).metadata().id

    assert first_metadata_id == second_metadata_id


# <https://github.com/delta-io/delta-rs/issues/2854>
@pytest.mark.polars
@pytest.mark.xfail(reason="polars needs update")
def test_write_binary_col_without_dssc(tmp_path: pathlib.Path):
    import polars as pl

    data_with_bin_col = {
        "x": [b"67890", b"abcde", b"xyzw", b"12345"],
        "y": [1, 2, 3, 4],
        "z": [101, 102, None, 104],
    }

    df_with_bin_col = pl.DataFrame(data_with_bin_col)
    df_with_bin_col.write_delta(tmp_path)

    assert len(df_with_bin_col.rows()) == 4

    dt = DeltaTable(tmp_path)
    stats = dt.get_add_actions(flatten=True)
    assert stats["null_count.x"].to_pylist() == [None]
    assert stats["min.x"].to_pylist() == [None]
    assert stats["max.x"].to_pylist() == [None]
    assert stats["null_count.y"].to_pylist() == [0]
    assert stats["min.y"].to_pylist() == [1]
    assert stats["max.y"].to_pylist() == [4]
    assert stats["null_count.z"].to_pylist() == [1]
    assert stats["min.z"].to_pylist() == [101]
    assert stats["max.z"].to_pylist() == [104]


# <https://github.com/delta-io/delta-rs/issues/2854>
@pytest.mark.polars
@pytest.mark.xfail(reason="polars needs update")
def test_write_binary_col_with_dssc(tmp_path: pathlib.Path):
    import polars as pl

    data_with_bin_col = {
        "x": [b"67890", b"abcde", b"xyzw", b"12345"],
        "y": [1, 2, 3, 4],
        "z.z": [101, 102, None, 104],
    }

    df_with_bin_col = pl.DataFrame(data_with_bin_col)
    df_with_bin_col.write_delta(
        tmp_path,
        delta_write_options={
            "configuration": {"delta.dataSkippingStatsColumns": "x,y,`z.z`"},
        },
    )

    assert len(df_with_bin_col.rows()) == 4

    dt = DeltaTable(tmp_path)
    stats = dt.get_add_actions(flatten=True)
    assert stats["null_count.x"].to_pylist() == [None]
    assert stats["min.x"].to_pylist() == [None]
    assert stats["max.x"].to_pylist() == [None]
    assert stats["null_count.y"].to_pylist() == [0]
    assert stats["min.y"].to_pylist() == [1]
    assert stats["max.y"].to_pylist() == [4]
    assert stats["null_count.z.z"].to_pylist() == [1]
    assert stats["min.z.z"].to_pylist() == [101]
    assert stats["max.z.z"].to_pylist() == [104]



from typing import Union

from deltalake._internal import (
    ArrayType,
    Field,
    MapType,
    PrimitiveType,
    Schema,
    StructType,
)

__all__ = ["ArrayType", "Field", "MapType", "PrimitiveType", "Schema", "StructType"]


# Can't implement inheritance (see note in src/schema.rs), so this is next
# best thing.
DataType = Union["PrimitiveType", "MapType", "StructType", "ArrayType"]



import pathlib
from typing import TYPE_CHECKING

import pytest

from deltalake import CommitProperties, convert_to_deltalake
from deltalake._internal import Field, PrimitiveType, Schema
from deltalake.exceptions import DeltaError
from deltalake.table import DeltaTable

if TYPE_CHECKING:
    import pyarrow as pa


@pytest.mark.pyarrow
def test_local_convert_to_delta(
    tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"
):
    import pyarrow.dataset as ds

    ds.write_dataset(
        sample_data_pyarrow,
        tmp_path,
        format="parquet",
        existing_data_behavior="overwrite_or_ignore",
    )

    name = "converted_table"
    description = "parquet table converted to delta table with delta-rs"
    convert_to_deltalake(
        tmp_path,
        name=name,
        description=description,
        configuration={"delta.appendOnly": "true"},
        commit_properties=CommitProperties(custom_metadata={"userName": "John Doe"}),
    )

    dt = DeltaTable(tmp_path)

    assert dt.version() == 0
    assert dt.files() == ["part-0.parquet"]
    assert dt.metadata().name == name
    assert dt.metadata().description == description
    assert dt.metadata().configuration == {"delta.appendOnly": "true"}
    assert dt.history()[0]["userName"] == "John Doe"


@pytest.mark.pyarrow
def test_convert_delta_write_modes(
    tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"
):
    import pyarrow.dataset as ds

    ds.write_dataset(
        sample_data_pyarrow,
        tmp_path,
        format="parquet",
        existing_data_behavior="overwrite_or_ignore",
    )

    convert_to_deltalake(
        tmp_path,
    )

    with pytest.raises(DeltaError):
        convert_to_deltalake(
            tmp_path,
        )

    convert_to_deltalake(tmp_path, mode="ignore")


@pytest.mark.pyarrow
def test_convert_delta_with_partitioning(
    tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"
):
    import pyarrow.dataset as ds

    ds.write_dataset(
        sample_data_pyarrow,
        tmp_path,
        format="parquet",
        existing_data_behavior="overwrite_or_ignore",
        partitioning=["utf8"],
        partitioning_flavor="hive",
    )

    schema = Schema(fields=[Field("utf8", PrimitiveType("string"))])

    with pytest.raises(
        DeltaError,
        match="Generic error: The schema of partition columns must be provided to convert a Parquet table to a Delta table",
    ):
        convert_to_deltalake(
            tmp_path,
        )
    with pytest.raises(
        ValueError, match="Partition strategy has to be provided with partition_by"
    ):
        convert_to_deltalake(
            tmp_path,
            partition_by=schema,
        )

    with pytest.raises(
        ValueError,
        match="Currently only `hive` partition strategy is supported to be converted.",
    ):
        convert_to_deltalake(
            tmp_path,
            partition_by=schema,
            partition_strategy="directory",  # type: ignore
        )

    convert_to_deltalake(
        tmp_path,
        partition_by=schema,
        partition_strategy="hive",
    )



from datetime import date, datetime
from typing import Any


def encode_partition_value(val: Any) -> str:
    # Rules based on: https://github.com/delta-io/delta/blob/master/PROTOCOL.md#partition-value-serialization
    if isinstance(val, bool):
        return str(val).lower()
    if isinstance(val, str):
        return val
    elif isinstance(val, (int, float)):
        return str(val)
    elif isinstance(val, datetime):
        return val.isoformat(sep=" ")
    elif isinstance(val, date):
        return val.isoformat()
    elif isinstance(val, bytes):
        return val.decode("unicode_escape", "backslashreplace")
    else:
        raise ValueError(f"Could not encode partition value for type: {val}")



from __future__ import annotations

from arro3.core import DataType
from arro3.core import Schema as Arro3Schema


def _convert_arro3_schema_to_delta(
    schema: Arro3Schema,
) -> Arro3Schema:
    """Convert a arro3 schema to a schema compatible with Delta Lake. Converts unsigned to signed equivalent, and
    converts all timestamps to `us` timestamps.
    Args
        schema: Source schema
    """

    def dtype_to_delta_dtype(dtype: DataType) -> DataType:
        # Handle nested types
        if (
            DataType.is_list(dtype)
            or DataType.is_large_list(dtype)
            or DataType.is_fixed_size_list(dtype)
            or DataType.is_list_view(dtype)
            or DataType.is_large_list_view(dtype)
        ):
            return list_to_delta_dtype(dtype)
        elif DataType.is_struct(dtype):
            return struct_to_delta_dtype(dtype)
        elif DataType.is_timestamp(dtype):
            if dtype.tz is None:
                return DataType.timestamp("us")
            else:
                return DataType.timestamp("us", tz="UTC")
        elif DataType.is_fixed_size_binary(dtype):
            return DataType.binary()
        elif DataType.is_unsigned_integer(dtype):
            if DataType.is_uint16(dtype):
                return DataType.int16()
            elif DataType.is_uint32(dtype):
                return DataType.int32()
            elif DataType.is_uint64(dtype):
                return DataType.int64()
            elif DataType.is_uint8(dtype):
                return DataType.int8()
            else:
                raise NotImplementedError
        else:
            return dtype

    def list_to_delta_dtype(
        dtype: DataType,
    ) -> DataType:
        nested_dtype = dtype.value_type
        inner_field = dtype.value_field

        assert nested_dtype is not None
        assert inner_field is not None

        inner_field_casted = inner_field.with_type(dtype_to_delta_dtype(nested_dtype))

        if DataType.is_large_list(dtype):
            return DataType.large_list(inner_field_casted)
        elif DataType.is_fixed_size_list(dtype):
            return DataType.list(inner_field_casted, dtype.list_size)
        elif DataType.is_large_list_view(dtype):
            return DataType.large_list_view(inner_field_casted)
        elif DataType.is_list_view(dtype):
            return DataType.list_view(inner_field_casted)
        elif DataType.is_list(dtype):
            return DataType.list(inner_field_casted)
        else:
            raise NotImplementedError

    def struct_to_delta_dtype(dtype: DataType) -> DataType:
        fields_cast = [f.with_type(dtype_to_delta_dtype(f.type)) for f in dtype.fields]
        return DataType.struct(fields_cast)

    return Arro3Schema([f.with_type(dtype_to_delta_dtype(f.type)) for f in schema])  # type: ignore[attr-defined]



from typing import Any, Callable

__version__: str
Schema: Any
Table: Any
RecordBatch: Any
RecordBatchReader: Any
Field: Any
DataType: Any
ListType: Any
StructType: Any
MapType: Any
FixedSizeListType: Any
LargeListViewType: Any
ListViewType: Any
FixedSizeBinaryType: Any
ExtensionType: Any
schema: Any
map_: Any
list_: Any
field: Any
struct: Any
type_for_alias: Any
date32: Any
date64: Any
decimal128: Any
int8: Any
int16: Any
int32: Any
int64: Any
uint8: Any
uint16: Any
uint32: Any
uint64: Any
float16: Any
float32: Any
float64: Any
large_string: Any
string: Any
large_binary: Any
binary: Any
binary_view: Any
string_view: Any
list_view: Any
large_list: Any
large_list_view: Any
LargeListType: Any
dictionary: Any
timestamp: Any
TimestampType: Any

py_buffer: Callable[[bytes], Any]
NativeFile: Any
BufferReader: Any
PythonFile: Any



"""Sphinx extension to add ReadTheDocs-style "Edit on GitHub" links to the
sidebar.

Loosely based on https://github.com/astropy/astropy/pull/347
"""

import os
import warnings

__licence__ = "BSD (3 clause)"


def get_github_url(app, view, path):
    return f"https://github.com/{app.config.edit_on_github_project}/{view}/{app.config.edit_on_github_branch}/{path}"


def html_page_context(app, pagename, templatename, context, doctree):
    if templatename != "page.html":
        return

    if not app.config.edit_on_github_project:
        warnings.warn("edit_on_github_project not specified")
        return

    path = os.path.relpath(doctree.get("source"), app.builder.srcdir)
    show_url = get_github_url(app, "blob", path)
    edit_url = get_github_url(app, "edit", path)

    context["show_on_github_url"] = show_url
    context["edit_on_github_url"] = edit_url

    # For sphinx_rtd_theme.
    context["display_github"] = True
    context["github_user"] = app.config.edit_on_github_project.split("/")[0]
    context["github_repo"] = app.config.edit_on_github_project.split("/")[1]
    context["github_version"] = (
        f"{app.config.edit_on_github_branch}/{app.config.page_source_prefix}/"
    )


def setup(app):
    app.add_config_value("edit_on_github_project", "", True)
    app.add_config_value("edit_on_github_branch", "main", True)
    app.add_config_value("page_source_prefix", "docs/source", True)
    app.connect("html-page-context", html_page_context)



from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Literal,
)

from deltalake._internal import Schema
from deltalake._internal import convert_to_deltalake as _convert_to_deltalake
from deltalake.writer._utils import try_get_deltatable

if TYPE_CHECKING:
    from deltalake.transaction import (
        CommitProperties,
        PostCommitHookProperties,
    )


def convert_to_deltalake(
    uri: str | Path,
    mode: Literal["error", "ignore"] = "error",
    partition_by: Schema | None = None,
    partition_strategy: Literal["hive"] | None = None,
    name: str | None = None,
    description: str | None = None,
    configuration: Mapping[str, str | None] | None = None,
    storage_options: dict[str, str] | None = None,
    commit_properties: CommitProperties | None = None,
    post_commithook_properties: PostCommitHookProperties | None = None,
) -> None:
    """
    `Convert` parquet tables `to delta` tables.

    Currently only HIVE partitioned tables are supported. `Convert to delta` creates
    a transaction log commit with add actions, and additional properties provided such
    as configuration, name, and description.

    Args:
        uri: URI of a table.
        partition_by: Optional partitioning schema if table is partitioned.
        partition_strategy: Optional partition strategy to read and convert
        mode: How to handle existing data. Default is to error if table already exists.
            If 'ignore', will not convert anything if table already exists.
        name: User-provided identifier for this table.
        description: User-provided description for this table.
        configuration: A map containing configuration options for the metadata action.
        storage_options: options passed to the native delta filesystem. Unused if 'filesystem' is defined.
        commit_properties: properties of the transaction commit. If None, default values are used.
        post_commithook_properties: properties for the post commit hook. If None, default values are used.
    """
    if partition_by is not None and partition_strategy is None:
        raise ValueError("Partition strategy has to be provided with partition_by.")

    if partition_strategy is not None and partition_strategy != "hive":
        raise ValueError(
            "Currently only `hive` partition strategy is supported to be converted."
        )

    if mode == "ignore" and try_get_deltatable(uri, storage_options) is not None:
        return

    _convert_to_deltalake(
        uri=str(uri),
        partition_schema=partition_by,
        partition_strategy=partition_strategy,
        name=name,
        description=description,
        configuration=configuration,
        storage_options=storage_options,
        commit_properties=commit_properties,
        post_commithook_properties=post_commithook_properties,
    )
    return



import pickle
import urllib
from typing import TYPE_CHECKING

import pytest

from deltalake import DeltaTable
from deltalake.exceptions import DeltaProtocolError
from deltalake.fs import DeltaStorageHandler
from deltalake.writer import write_deltalake

if TYPE_CHECKING:
    import pyarrow as pa

try:
    import pyarrow as pa
    import pyarrow.parquet as pq
    from pyarrow.fs import FileType
except ImportError:
    pass


@pytest.mark.pyarrow
@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=15, method="thread")
def test_read_files(s3_localstack):
    table_path = "s3://deltars/simple"
    handler = DeltaStorageHandler(table_path)
    dt = DeltaTable(table_path)
    files = dt.file_uris()
    assert len(files) > 0
    for file in files:
        rel_path = file[len(table_path) :]
        with handler.open_input_file(rel_path) as f_:
            table = pq.read_table(f_)
            assert isinstance(table, pa.Table)
            assert table.shape > (0, 0)


@pytest.mark.pyarrow
@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=15, method="thread")
def test_read_file_info(s3_localstack):
    table_path = "s3://deltars/simple"
    handler = DeltaStorageHandler(table_path)
    meta = handler.get_file_info(
        ["part-00000-a72b1fb3-f2df-41fe-a8f0-e65b746382dd-c000.snappy.parquet"]
    )
    assert len(meta) == 1
    assert meta[0].type == FileType.File


@pytest.mark.pyarrow
@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=15, method="thread")
def test_s3_authenticated_read_write(s3_localstack_creds, monkeypatch):
    monkeypatch.setenv("AWS_DEFAULT_REGION", "us-east-1")
    # Create unauthenticated handler
    storage_handler = DeltaStorageHandler(
        "s3://deltars/",
        options={
            "AWS_ENDPOINT_URL": s3_localstack_creds["AWS_ENDPOINT_URL"],
            # Grants anonymous access. If we don't do this, will timeout trying
            # to reading from EC2 instance provider.
            "AWS_ACCESS_KEY_ID": "",
            "AWS_SECRET_ACCESS_KEY": "",
        },
    )

    # Make a get request on an object
    with pytest.raises(Exception):
        storage_handler.open_input_stream("nonexistent")

    # Try to write an object
    with pytest.raises(Exception):
        storage_handler.open_output_stream("nonexistent")


@pytest.mark.pyarrow
@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=15, method="thread")
def test_read_simple_table_from_remote(s3_localstack):
    table_path = "s3://deltars/simple"
    dt = DeltaTable(table_path)
    assert dt.to_pyarrow_table().equals(pa.table({"id": [5, 7, 9]}))

    expected_files = [
        "part-00000-2befed33-c358-4768-a43c-3eda0d2a499d-c000.snappy.parquet",
        "part-00000-c1777d7d-89d9-4790-b38a-6ee7e24456b1-c000.snappy.parquet",
        "part-00001-7891c33d-cedc-47c3-88a6-abcfb049d3b4-c000.snappy.parquet",
        "part-00004-315835fe-fb44-4562-98f6-5e6cfa3ae45d-c000.snappy.parquet",
        "part-00007-3a0e4727-de0d-41b6-81ef-5223cf40f025-c000.snappy.parquet",
    ]

    assert dt.files() == expected_files
    assert dt.file_uris() == [table_path + "/" + path for path in expected_files]


@pytest.mark.pyarrow
@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=15, method="thread")
@pytest.mark.skip(
    reason="Temporarily disabled until we can resolve https://github.com/delta-io/delta-rs/pull/2120#issuecomment-1912367573"
)
def test_roundtrip_s3_env(s3_localstack, sample_data_pyarrow: "pa.Table", monkeypatch):
    table_path = "s3://deltars/roundtrip"

    # Create new table with path
    with pytest.raises(DeltaProtocolError, match="Atomic rename requires a LockClient"):
        write_deltalake(table_path, sample_data_pyarrow)

    monkeypatch.setenv("AWS_S3_ALLOW_UNSAFE_RENAME", "true")

    # Create new table with path
    write_deltalake(table_path, sample_data_pyarrow)
    dt = DeltaTable(table_path)
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow
    assert dt.version() == 0

    # Write with existing DeltaTable
    write_deltalake(dt, sample_data_pyarrow, mode="overwrite")
    dt.update_incremental()
    assert dt.version() == 1

    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow


@pytest.mark.pyarrow
@pytest.mark.s3
@pytest.mark.integration
@pytest.mark.timeout(timeout=15, method="thread")
def test_roundtrip_s3_direct(s3_localstack_creds, sample_data_pyarrow: "pa.Table"):
    table_path = "s3://deltars/roundtrip2"

    # Fails without any credentials
    with pytest.raises(IOError):
        anon_storage_options = {
            "AWS_ENDPOINT_URL": s3_localstack_creds["AWS_ENDPOINT_URL"],
            # Grants anonymous access. If we don't do this, will timeout trying
            # to reading from EC2 instance provider.
            "AWS_ACCESS_KEY_ID": "",
            "AWS_SECRET_ACCESS_KEY": "",
        }
        write_deltalake(
            table_path,
            sample_data_pyarrow,
            storage_options=anon_storage_options,
        )

    # Can pass storage_options in directly
    storage_opts = {
        "AWS_ALLOW_HTTP": "true",
        "AWS_S3_ALLOW_UNSAFE_RENAME": "true",
    }
    storage_opts.update(s3_localstack_creds)
    write_deltalake(table_path, sample_data_pyarrow, storage_options=storage_opts)
    dt = DeltaTable(table_path, storage_options=storage_opts)
    assert dt.version() == 0
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow

    # Can pass storage_options into DeltaTable and then write
    write_deltalake(dt, sample_data_pyarrow, mode="overwrite")
    dt.update_incremental()
    assert dt.version() == 1
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow


@pytest.mark.pyarrow
@pytest.mark.azure
@pytest.mark.integration
@pytest.mark.timeout(timeout=60, method="thread")
def test_roundtrip_azure_env(azurite_env_vars, sample_data_pyarrow: "pa.Table"):
    table_path = "az://deltars/roundtrip"

    # Create new table with path
    write_deltalake(table_path, sample_data_pyarrow)
    dt = DeltaTable(table_path)
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow
    assert dt.version() == 0

    # Write with existing DeltaTable
    write_deltalake(dt, sample_data_pyarrow, mode="overwrite")
    dt.update_incremental()
    assert dt.version() == 1

    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow


@pytest.mark.pyarrow
@pytest.mark.azure
@pytest.mark.integration
@pytest.mark.timeout(timeout=60, method="thread")
def test_roundtrip_azure_direct(azurite_creds, sample_data_pyarrow: "pa.Table"):
    table_path = "az://deltars/roundtrip2"

    # Can pass storage_options in directly
    write_deltalake(table_path, sample_data_pyarrow, storage_options=azurite_creds)
    dt = DeltaTable(table_path, storage_options=azurite_creds)
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow
    assert dt.version() == 0

    # Can pass storage_options into DeltaTable and then write
    write_deltalake(dt, sample_data_pyarrow, mode="overwrite")
    dt.update_incremental()
    assert dt.version() == 1

    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow


@pytest.mark.pyarrow
@pytest.mark.azure
@pytest.mark.integration
@pytest.mark.timeout(timeout=60, method="thread")
def test_roundtrip_azure_sas(azurite_sas_creds, sample_data_pyarrow: "pa.Table"):
    table_path = "az://deltars/roundtrip3"
    write_deltalake(table_path, sample_data_pyarrow, storage_options=azurite_sas_creds)
    dt = DeltaTable(table_path, storage_options=azurite_sas_creds)
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow
    assert dt.version() == 0


@pytest.mark.pyarrow
@pytest.mark.azure
@pytest.mark.integration
@pytest.mark.timeout(timeout=60, method="thread")
def test_roundtrip_azure_decoded_sas(
    azurite_sas_creds, sample_data_pyarrow: "pa.Table"
):
    table_path = "az://deltars/roundtrip4"
    azurite_sas_creds["SAS_TOKEN"] = urllib.parse.unquote(
        azurite_sas_creds["SAS_TOKEN"]
    )

    write_deltalake(table_path, sample_data_pyarrow, storage_options=azurite_sas_creds)
    dt = DeltaTable(table_path, storage_options=azurite_sas_creds)
    table = dt.to_pyarrow_table()
    assert table == sample_data_pyarrow
    assert dt.version() == 0


@pytest.mark.pyarrow
@pytest.mark.parametrize("storage_size", [1, 4 * 1024 * 1024, 5 * 1024 * 1024 - 1])
def test_warning_for_small_max_buffer_size(tmp_path, storage_size):
    storage_opts = {"max_buffer_size": str(storage_size)}
    store = DeltaStorageHandler(str(tmp_path.absolute()), options=storage_opts)
    with pytest.warns(UserWarning) as warnings:
        store.open_output_stream("test")

    assert len(warnings) == 1
    assert (
        f"You specified a `max_buffer_size` of {storage_size} bits less than {5 * 1024 * 1024} bits"
        in str(warnings[0].message)
    )


@pytest.mark.pyarrow
def test_pickle_roundtrip(tmp_path):
    store = DeltaStorageHandler(str(tmp_path.absolute()))

    with (tmp_path / "asd.pkl").open("wb") as handle:
        pickle.dump(store, handle)

    with (tmp_path / "asd.pkl").open("rb") as handle:
        store_pkl = pickle.load(handle)

    infos = store_pkl.get_file_info(["asd.pkl"])
    assert infos[0].size > 0



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Literal


class Compression(Enum):
    UNCOMPRESSED = "UNCOMPRESSED"
    SNAPPY = "SNAPPY"
    GZIP = "GZIP"
    BROTLI = "BROTLI"
    LZ4 = "LZ4"
    ZSTD = "ZSTD"
    LZ4_RAW = "LZ4_RAW"

    @classmethod
    def from_str(cls, value: str) -> Compression:
        try:
            return cls(value.upper())
        except ValueError:
            raise ValueError(
                f"{value} is not a valid Compression. Valid values are: {[item.value for item in Compression]}"
            )

    def get_level_range(self) -> tuple[int, int]:
        if self == Compression.GZIP:
            MIN_LEVEL = 0
            MAX_LEVEL = 10
        elif self == Compression.BROTLI:
            MIN_LEVEL = 0
            MAX_LEVEL = 11
        elif self == Compression.ZSTD:
            MIN_LEVEL = 1
            MAX_LEVEL = 22
        else:
            raise KeyError(f"{self.value} does not have a compression level.")
        return MIN_LEVEL, MAX_LEVEL

    def get_default_level(self) -> int:
        if self == Compression.GZIP:
            DEFAULT = 6
        elif self == Compression.BROTLI:
            DEFAULT = 1
        elif self == Compression.ZSTD:
            DEFAULT = 1
        else:
            raise KeyError(f"{self.value} does not have a compression level.")
        return DEFAULT

    def check_valid_level(self, level: int) -> bool:
        MIN_LEVEL, MAX_LEVEL = self.get_level_range()
        if level < MIN_LEVEL or level > MAX_LEVEL:
            raise ValueError(
                f"Compression level for {self.value} should fall between {MIN_LEVEL}-{MAX_LEVEL}"
            )
        else:
            return True


@dataclass(init=True)
class BloomFilterProperties:
    """The Bloom Filter Properties instance for the Rust parquet writer."""

    def __init__(
        self,
        set_bloom_filter_enabled: bool | None,
        fpp: float | None = None,
        ndv: int | None = None,
    ) -> None:
        """Create a Bloom Filter Properties instance for the Rust parquet writer:

        Args:
            set_bloom_filter_enabled: If True and no fpp or ndv are provided, the default values will be used.
            fpp: The false positive probability for the bloom filter. Must be between 0 and 1 exclusive.
            ndv: The number of distinct values for the bloom filter.
        """
        if fpp is not None and (fpp <= 0 or fpp >= 1):
            raise ValueError("fpp must be between 0 and 1 exclusive")
        self.set_bloom_filter_enabled = set_bloom_filter_enabled
        self.fpp = fpp
        self.ndv = ndv

    def __str__(self) -> str:
        return f"set_bloom_filter_enabled: {self.set_bloom_filter_enabled}, fpp: {self.fpp}, ndv: {self.ndv}"


@dataclass(init=True)
class ColumnProperties:
    """The Column Properties instance for the Rust parquet writer."""

    def __init__(
        self,
        dictionary_enabled: bool | None = None,
        statistics_enabled: Literal["NONE", "CHUNK", "PAGE"] | None = None,
        bloom_filter_properties: BloomFilterProperties | None = None,
    ) -> None:
        """Create a Column Properties instance for the Rust parquet writer:

        Args:
            dictionary_enabled: Enable dictionary encoding for the column.
            statistics_enabled: Statistics level for the column.
            bloom_filter_properties: Bloom Filter Properties for the column.
        """
        self.dictionary_enabled = dictionary_enabled
        self.statistics_enabled = statistics_enabled
        self.bloom_filter_properties = bloom_filter_properties

    def __str__(self) -> str:
        return (
            f"dictionary_enabled: {self.dictionary_enabled}, statistics_enabled: {self.statistics_enabled}, "
            f"bloom_filter_properties: {self.bloom_filter_properties}"
        )


@dataclass(init=True)
class WriterProperties:
    """A Writer Properties instance for the Rust parquet writer."""

    def __init__(
        self,
        data_page_size_limit: int | None = None,
        dictionary_page_size_limit: int | None = None,
        data_page_row_count_limit: int | None = None,
        write_batch_size: int | None = None,
        max_row_group_size: int | None = None,
        compression: Literal[
            "UNCOMPRESSED", "SNAPPY", "GZIP", "BROTLI", "LZ4", "ZSTD", "LZ4_RAW"
        ]
        | None = None,
        compression_level: int | None = None,
        statistics_truncate_length: int | None = None,
        default_column_properties: ColumnProperties | None = None,
        column_properties: dict[str, ColumnProperties] | None = None,
    ) -> None:
        """Create a Writer Properties instance for the Rust parquet writer:

        Args:
            data_page_size_limit: Limit DataPage size to this in bytes.
            dictionary_page_size_limit: Limit the size of each DataPage to store dicts to this amount in bytes.
            data_page_row_count_limit: Limit the number of rows in each DataPage.
            write_batch_size: Splits internally to smaller batch size.
            max_row_group_size: Max number of rows in row group.
            compression: compression type.
            compression_level: If none and compression has a level, the default level will be used, only relevant for
                GZIP: levels (1-9),
                BROTLI: levels (1-11),
                ZSTD: levels (1-22),
            statistics_truncate_length: maximum length of truncated min/max values in statistics.
            default_column_properties: Default Column Properties for the Rust parquet writer.
            column_properties: Column Properties for the Rust parquet writer.
        """
        self.data_page_size_limit = data_page_size_limit
        self.dictionary_page_size_limit = dictionary_page_size_limit
        self.data_page_row_count_limit = data_page_row_count_limit
        self.write_batch_size = write_batch_size
        self.max_row_group_size = max_row_group_size
        self.compression = None
        self.statistics_truncate_length = statistics_truncate_length
        self.default_column_properties = default_column_properties
        self.column_properties = column_properties

        if compression_level is not None and compression is None:
            raise ValueError(
                """Providing a compression level without the compression type is not possible,
                             please provide the compression as well."""
            )
        if isinstance(compression, str):
            compression_enum = Compression.from_str(compression)
            if compression_enum in [
                Compression.GZIP,
                Compression.BROTLI,
                Compression.ZSTD,
            ]:
                if compression_level is not None:
                    if compression_enum.check_valid_level(compression_level):
                        parquet_compression = (
                            f"{compression_enum.value}({compression_level})"
                        )
                else:
                    parquet_compression = f"{compression_enum.value}({compression_enum.get_default_level()})"
            else:
                parquet_compression = compression_enum.value
            self.compression = parquet_compression

    def __str__(self) -> str:
        column_properties_str = (
            ", ".join([f"column '{k}': {v}" for k, v in self.column_properties.items()])
            if self.column_properties
            else None
        )
        return (
            f"WriterProperties(data_page_size_limit: {self.data_page_size_limit}, dictionary_page_size_limit: {self.dictionary_page_size_limit}, "
            f"data_page_row_count_limit: {self.data_page_row_count_limit}, write_batch_size: {self.write_batch_size}, "
            f"max_row_group_size: {self.max_row_group_size}, compression: {self.compression}, statistics_truncate_length: {self.statistics_truncate_length},"
            f"default_column_properties: {self.default_column_properties}, column_properties: {column_properties_str})"
        )



from deltalake.writer.convert_to import convert_to_deltalake
from deltalake.writer.properties import (
    BloomFilterProperties,
    ColumnProperties,
    WriterProperties,
)
from deltalake.writer.writer import write_deltalake

__all__ = [
    "BloomFilterProperties",
    "ColumnProperties",
    "WriterProperties",
    "convert_to_deltalake",
    "write_deltalake",
]



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from deltalake.exceptions import TableNotFoundError

if TYPE_CHECKING:
    from deltalake.table import DeltaTable


def try_get_table_and_table_uri(
    table_or_uri: str | Path | DeltaTable,
    storage_options: dict[str, str] | None = None,
) -> tuple[DeltaTable | None, str]:
    """Parses the `table_or_uri`.
    Raises [ValueError] If `table_or_uri` is not of type str, Path or DeltaTable.

    Args:
        table_or_uri: URI of a table or a DeltaTable object.
        storage_options: Options passed to the native delta filesystem.

    Returns:
        (DeltaTable object, URI of the table)
    """
    from deltalake.table import DeltaTable

    if not isinstance(table_or_uri, (str, Path, DeltaTable)):
        raise ValueError("table_or_uri must be a str, Path or DeltaTable")

    if isinstance(table_or_uri, (str, Path)):
        table = try_get_deltatable(table_or_uri, storage_options)
        table_uri = str(table_or_uri)
    else:
        table = table_or_uri
        table_uri = table._table.table_uri()

    return (table, table_uri)


def try_get_deltatable(
    table_uri: str | Path, storage_options: dict[str, str] | None
) -> DeltaTable | None:
    from deltalake.table import DeltaTable

    try:
        return DeltaTable(table_uri, storage_options=storage_options)
    except TableNotFoundError:
        return None



import datetime as dt
import os
import pathlib
import shutil
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING

import pytest
from arro3.core import Array, DataType, Table
from arro3.core import Field as ArrowField

from deltalake import (
    DeltaTable,
    PostCommitHookProperties,
    QueryBuilder,
    write_deltalake,
)

if TYPE_CHECKING:
    import pyarrow as pa


def test_checkpoint(tmp_path: pathlib.Path, sample_table: Table):
    tmp_table_path = tmp_path / "path" / "to" / "table"
    checkpoint_path = tmp_table_path / "_delta_log" / "_last_checkpoint"
    last_checkpoint_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.checkpoint.parquet"
    )
    write_deltalake(str(tmp_table_path), sample_table)

    assert not checkpoint_path.exists()

    delta_table = DeltaTable(str(tmp_table_path))
    delta_table.create_checkpoint()

    assert last_checkpoint_path.exists()
    assert checkpoint_path.exists()


def setup_cleanup_metadata(tmp_path: pathlib.Path, sample_table: Table):
    tmp_table_path = tmp_path / "path" / "to" / "table"
    first_log_path = tmp_table_path / "_delta_log" / "00000000000000000000.json"
    first_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.json.tmp"
    )
    second_log_path = tmp_table_path / "_delta_log" / "00000000000000000001.json"
    second_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000002.json.tmp"
    )
    third_log_path = tmp_table_path / "_delta_log" / "00000000000000000002.json"

    # Create few log files
    write_deltalake(str(tmp_table_path), sample_table)
    write_deltalake(str(tmp_table_path), sample_table, mode="overwrite")
    delta_table = DeltaTable(str(tmp_table_path))
    delta_table.delete()

    # Create failed json commit
    shutil.copy(str(first_log_path), str(first_failed_log_path))
    shutil.copy(str(third_log_path), str(second_failed_log_path))

    # Move first failed log entry timestamp back in time for more than 30 days
    old_ts = (dt.datetime.now() - dt.timedelta(days=31)).timestamp()
    os.utime(first_failed_log_path, (old_ts, old_ts))

    # Move first log entry timestamp back in time for more than 30 days
    old_ts = (dt.datetime.now() - dt.timedelta(days=31)).timestamp()
    os.utime(first_log_path, (old_ts, old_ts))

    # Move second log entry timestamp back in time for a minute
    near_ts = (dt.datetime.now() - dt.timedelta(minutes=1)).timestamp()
    os.utime(second_log_path, (near_ts, near_ts))

    assert first_log_path.exists()
    assert first_failed_log_path.exists()
    assert second_log_path.exists()
    assert third_log_path.exists()
    assert second_failed_log_path.exists()
    return delta_table


def test_cleanup_metadata(tmp_path: pathlib.Path, sample_table: Table):
    delta_table = setup_cleanup_metadata(tmp_path, sample_table)
    delta_table.create_checkpoint()
    delta_table.cleanup_metadata()

    tmp_table_path = tmp_path / "path" / "to" / "table"
    first_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.json.tmp"
    )
    first_log_path = tmp_table_path / "_delta_log" / "00000000000000000000.json"
    second_log_path = tmp_table_path / "_delta_log" / "00000000000000000001.json"
    second_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000002.json.tmp"
    )
    third_log_path = tmp_table_path / "_delta_log" / "00000000000000000002.json"

    assert not first_log_path.exists()
    assert not first_failed_log_path.exists()
    assert second_log_path.exists()
    assert third_log_path.exists()
    assert second_failed_log_path.exists()


def test_cleanup_metadata_log_cleanup_hook(tmp_path: pathlib.Path, sample_table: Table):
    delta_table = setup_cleanup_metadata(tmp_path, sample_table)
    delta_table.create_checkpoint()

    write_deltalake(delta_table, sample_table, mode="append")

    tmp_table_path = tmp_path / "path" / "to" / "table"
    first_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.json.tmp"
    )
    first_log_path = tmp_table_path / "_delta_log" / "00000000000000000000.json"
    second_log_path = tmp_table_path / "_delta_log" / "00000000000000000001.json"
    second_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000002.json.tmp"
    )
    third_log_path = tmp_table_path / "_delta_log" / "00000000000000000002.json"

    assert not first_log_path.exists()
    assert not first_failed_log_path.exists()
    assert second_log_path.exists()
    assert third_log_path.exists()
    assert second_failed_log_path.exists()


def test_cleanup_metadata_log_cleanup_hook_disabled(
    tmp_path: pathlib.Path, sample_table: Table
):
    delta_table = setup_cleanup_metadata(tmp_path, sample_table)
    delta_table.create_checkpoint()

    write_deltalake(
        delta_table,
        sample_table,
        mode="append",
        post_commithook_properties=PostCommitHookProperties(cleanup_expired_logs=False),
    )

    tmp_table_path = tmp_path / "path" / "to" / "table"
    first_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.json.tmp"
    )
    first_log_path = tmp_table_path / "_delta_log" / "00000000000000000000.json"
    second_log_path = tmp_table_path / "_delta_log" / "00000000000000000001.json"
    second_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000002.json.tmp"
    )
    third_log_path = tmp_table_path / "_delta_log" / "00000000000000000002.json"

    assert first_log_path.exists()
    assert first_failed_log_path.exists()
    assert second_log_path.exists()
    assert third_log_path.exists()
    assert second_failed_log_path.exists()


def test_cleanup_metadata_no_checkpoint(tmp_path: pathlib.Path, sample_table: Table):
    delta_table = setup_cleanup_metadata(tmp_path, sample_table)
    delta_table.cleanup_metadata()

    tmp_table_path = tmp_path / "path" / "to" / "table"
    first_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.json.tmp"
    )
    first_log_path = tmp_table_path / "_delta_log" / "00000000000000000000.json"
    second_log_path = tmp_table_path / "_delta_log" / "00000000000000000001.json"
    second_failed_log_path = (
        tmp_table_path / "_delta_log" / "00000000000000000002.json.tmp"
    )
    third_log_path = tmp_table_path / "_delta_log" / "00000000000000000002.json"

    assert first_log_path.exists()
    assert first_failed_log_path.exists()
    assert second_log_path.exists()
    assert third_log_path.exists()
    assert second_failed_log_path.exists()


@pytest.mark.pyarrow
def test_features_maintained_after_checkpoint(tmp_path: pathlib.Path):
    from datetime import datetime

    import pyarrow as pa

    data = pa.table(
        {
            "timestamp": pa.array([datetime(2022, 1, 1)]),
        }
    )
    write_deltalake(tmp_path, data)

    dt = DeltaTable(tmp_path)
    current_protocol = dt.protocol()

    dt.create_checkpoint()

    dt = DeltaTable(tmp_path)
    protocol_after_checkpoint = dt.protocol()

    assert protocol_after_checkpoint.reader_features == ["timestampNtz"]
    assert current_protocol == protocol_after_checkpoint


@pytest.mark.pyarrow
def test_features_null_on_below_v3_v7(tmp_path: pathlib.Path):
    import pyarrow as pa
    import pyarrow.parquet as pq

    data = pa.table(
        {
            "int": pa.array([1]),
        }
    )
    write_deltalake(tmp_path, data)

    dt = DeltaTable(tmp_path)
    current_protocol = dt.protocol()

    dt.create_checkpoint()

    dt = DeltaTable(tmp_path)
    protocol_after_checkpoint = dt.protocol()

    assert protocol_after_checkpoint.reader_features is None
    assert protocol_after_checkpoint.writer_features is None
    assert current_protocol == protocol_after_checkpoint

    checkpoint = pq.read_table(
        os.path.join(tmp_path, "_delta_log/00000000000000000000.checkpoint.parquet")
    )

    assert checkpoint["protocol"][0]["writerFeatures"].as_py() is None
    assert checkpoint["protocol"][0]["readerFeatures"].as_py() is None


@pytest.fixture
def sample_all_types():
    from datetime import timezone

    import pyarrow as pa

    nrows = 5
    return pa.table(
        {
            "utf8": pa.array([str(x) for x in range(nrows)]),
            "int64": pa.array(list(range(nrows)), pa.int64()),
            "int32": pa.array(list(range(nrows)), pa.int32()),
            "int16": pa.array(list(range(nrows)), pa.int16()),
            "int8": pa.array(list(range(nrows)), pa.int8()),
            "float32": pa.array([float(x) for x in range(nrows)], pa.float32()),
            "float64": pa.array([float(x) for x in range(nrows)], pa.float64()),
            "bool": pa.array([x % 2 == 0 for x in range(nrows)]),
            "binary": pa.array([str(x).encode() for x in range(nrows)]),
            # "decimal": pa.array([Decimal("10.000") + x for x in range(nrows)]), # Some issue with decimal and Rust engine at the moment.
            "date32": pa.array(
                [date(2022, 1, 1) + timedelta(days=x) for x in range(nrows)]
            ),
            "timestampNtz": pa.array(
                [datetime(2022, 1, 1) + timedelta(hours=x) for x in range(nrows)]
            ),
            "timestamp": pa.array(
                [
                    datetime(2022, 1, 1, tzinfo=timezone.utc) + timedelta(hours=x)
                    for x in range(nrows)
                ]
            ),
            "struct": pa.array([{"x": x, "y": str(x)} for x in range(nrows)]),
            "list": pa.array([list(range(x + 1)) for x in range(nrows)]),
        }
    )


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "part_col",
    [
        "timestampNtz",
        "timestamp",
    ],
)
def test_checkpoint_partition_timestamp_2380(
    tmp_path: pathlib.Path, sample_all_types: "pa.Table", part_col: str
):
    tmp_table_path = tmp_path / "path" / "to" / "table"
    checkpoint_path = tmp_table_path / "_delta_log" / "_last_checkpoint"
    last_checkpoint_path = (
        tmp_table_path / "_delta_log" / "00000000000000000000.checkpoint.parquet"
    )

    # TODO: Include binary after fixing issue "Json error: binary type is not supported"
    sample_data_pyarrow = sample_all_types.drop(["binary"])
    write_deltalake(
        str(tmp_table_path),
        sample_data_pyarrow,
        partition_by=[part_col],
    )

    assert not checkpoint_path.exists()

    delta_table = DeltaTable(str(tmp_table_path))

    delta_table.create_checkpoint()

    assert last_checkpoint_path.exists()
    assert checkpoint_path.exists()


def test_checkpoint_with_binary_column(tmp_path: pathlib.Path):
    data = Table(
        {
            "intColumn": Array(
                [1],
                ArrowField("intColumn", type=DataType.int64(), nullable=True),
            ),
            "binaryColumn": Array(
                [b"a"],
                ArrowField("binaryColumn", type=DataType.binary(), nullable=True),
            ),
        }
    )

    write_deltalake(
        str(tmp_path),
        data,
        partition_by=["intColumn"],
        mode="append",
    )

    dt = DeltaTable(tmp_path)
    dt.create_checkpoint()

    dt = DeltaTable(tmp_path)

    assert (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select intColumn, binaryColumn from tbl")
        .read_all()
        == data
    )


@pytest.mark.pyarrow
def test_checkpoint_post_commit_config(
    tmp_path: pathlib.Path, sample_data_pyarrow: "pa.Table"
):
    """Checks whether checkpoints are properly written based on commit_interval"""
    tmp_table_path = tmp_path / "path" / "to" / "table"
    checkpoint_path = tmp_table_path / "_delta_log" / "_last_checkpoint"
    first_checkpoint_path = (
        tmp_table_path / "_delta_log" / "00000000000000000004.checkpoint.parquet"
    )
    second_checkpoint_path = (
        tmp_table_path / "_delta_log" / "00000000000000000009.checkpoint.parquet"
    )

    # TODO: Include binary after fixing issue "Json error: binary type is not supported"
    sample_data_pyarrow = sample_data_pyarrow.drop(["binary"])
    for i in range(2):
        write_deltalake(
            str(tmp_table_path),
            sample_data_pyarrow,
            mode="append",
            configuration={"delta.checkpointInterval": "5"},
        )

    assert not checkpoint_path.exists()
    assert not first_checkpoint_path.exists()
    assert not second_checkpoint_path.exists()

    for i in range(10):
        write_deltalake(
            str(tmp_table_path),
            sample_data_pyarrow,
            mode="append",
            configuration={"delta.checkpointInterval": "5"},
        )

    assert checkpoint_path.exists()
    assert first_checkpoint_path.exists()
    assert second_checkpoint_path.exists()

    for i in range(12):
        if i in [4, 9]:
            continue
        random_checkpoint_path = (
            tmp_table_path / "_delta_log" / f"{str(i).zfill(20)}.checkpoint.parquet"
        )
        assert not random_checkpoint_path.exists()

    dt = DeltaTable(str(tmp_table_path))
    assert dt.version() == 11


def test_checkpoint_post_commit_config_multiple_operations(
    tmp_path: pathlib.Path, sample_table: Table
):
    """Checks whether checkpoints are properly written based on commit_interval"""
    tmp_table_path = tmp_path / "path" / "to" / "table"
    checkpoint_path = tmp_table_path / "_delta_log" / "_last_checkpoint"
    first_checkpoint_path = (
        tmp_table_path / "_delta_log" / "00000000000000000004.checkpoint.parquet"
    )
    second_checkpoint_path = (
        tmp_table_path / "_delta_log" / "00000000000000000009.checkpoint.parquet"
    )

    for i in range(4):
        write_deltalake(
            str(tmp_table_path),
            sample_table,
            mode="append",
            configuration={"delta.checkpointInterval": "5"},
        )

    assert not checkpoint_path.exists()
    assert not first_checkpoint_path.exists()
    assert not second_checkpoint_path.exists()

    dt = DeltaTable(str(tmp_table_path))

    dt.optimize.compact()

    assert checkpoint_path.exists()
    assert first_checkpoint_path.exists()

    for i in range(4):
        write_deltalake(
            str(tmp_table_path),
            sample_table,
            mode="append",
            configuration={"delta.checkpointInterval": "5"},
        )

    dt = DeltaTable(str(tmp_table_path))
    dt.delete()

    assert second_checkpoint_path.exists()

    for i in range(12):
        if i in [4, 9]:
            continue
        random_checkpoint_path = (
            tmp_table_path / "_delta_log" / f"{str(i).zfill(20)}.checkpoint.parquet"
        )
        assert not random_checkpoint_path.exists()

    delta_table = DeltaTable(str(tmp_table_path))
    assert delta_table.version() == 9


@pytest.mark.pyarrow
def test_checkpoint_with_nullable_false(tmp_path: pathlib.Path):
    tmp_table_path = tmp_path / "path" / "to" / "table"
    checkpoint_path = tmp_table_path / "_delta_log" / "_last_checkpoint"
    import pyarrow as pa

    pylist = [{"year": 2023, "n_party": 0}, {"year": 2024, "n_party": 1}]
    my_schema = pa.schema(
        [
            pa.field("year", pa.int64(), nullable=False),
            pa.field("n_party", pa.int64(), nullable=False),
        ]
    )

    data = pa.Table.from_pylist(pylist, schema=my_schema)

    write_deltalake(
        str(tmp_table_path),
        data,
        configuration={"delta.dataSkippingNumIndexedCols": "1"},
    )

    DeltaTable(str(tmp_table_path)).create_checkpoint()

    assert checkpoint_path.exists()

    assert DeltaTable(str(tmp_table_path)).to_pyarrow_table() == data


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_checkpoint_with_multiple_writes(tmp_path: pathlib.Path):
    import pandas as pd

    write_deltalake(
        tmp_path,
        pd.DataFrame(
            {
                "a": ["a"],
                "b": [3],
            }
        ),
    )
    dt = DeltaTable(tmp_path)
    dt.create_checkpoint()
    assert dt.version() == 0
    df = pd.DataFrame(
        {
            "a": ["a"],
            "b": [100],
        }
    )
    write_deltalake(tmp_path, df, mode="overwrite")

    dt = DeltaTable(tmp_path)
    assert dt.version() == 1
    new_df = dt.to_pandas()
    print(dt.to_pandas())
    assert len(new_df) == 1, "We overwrote! there should only be one row"


@pytest.mark.polars
@pytest.mark.xfail(reason="polars needs update")
def test_refresh_snapshot_after_log_cleanup_3057(tmp_path):
    """https://github.com/delta-io/delta-rs/issues/3057"""
    import polars as pl

    configuration = {
        "delta.deletedFileRetentionDuration": "interval 0 days",
        "delta.logRetentionDuration": "interval 0 days",
        "delta.targetFileSize": str(128 * 1024 * 1024),
    }

    for i in range(2):
        df = pl.DataFrame({"foo": [i]})
        df.write_delta(
            str(tmp_path),
            delta_write_options={"configuration": configuration},
            mode="append",
        )

    # create checkpoint so that logs before checkpoint can get removed
    dt = DeltaTable(tmp_path)
    dt.create_checkpoint()

    # Write to table again, snapshot should be correctly refreshed so that clean_up metadata can run after this
    df = pl.DataFrame({"foo": [1]})
    df.write_delta(dt, mode="append")

    # Vacuum is noop, since we already removed logs before and snapshot doesn't reference them anymore

    vacuum_log = dt.vacuum(
        retention_hours=0, enforce_retention_duration=False, dry_run=False
    )

    assert vacuum_log == []
